import os
from typing import TYPE_CHECKING, Union
from PyQt5.QtXml import QDomDocument, QDomElement
from pathlib import Path

import ray

if TYPE_CHECKING:
    from session import Session
    from osc_server_thread import OscServerThread

instance = None

class Daemon:
    net_daemon_id = 0
    root = ""
    session_path = ""
    pid = 0
    port = 0
    user = ""
    not_default = False


class MultiDaemonFile:
    file_path = '/tmp/RaySession/multi-daemon.xml'

    def __init__(self, session: 'Session', server: 'OscServerThread'):
        self.session = session
        self.server = server

        self._xml = QDomDocument()

        global instance
        instance = self

        self._locked_session_paths = set()

    @staticmethod
    def get_instance():
        return instance

    def _pid_exists(self, pid)->bool:
        if isinstance(pid, str):
            pid = int(pid)

        try:
            os.kill(pid, 0)
        except OSError:
            return False
        else:
            return True

    def _remove_file(self):
        try:
            os.remove(self.file_path)
        except:
            return

    def _open_file(self)->bool:
        if not os.path.exists(self.file_path):
            dir_path = os.path.dirname(self.file_path)
            if not os.path.exists(dir_path):
                os.makedirs(dir_path)
                # give read/write access for all users
                os.chmod(dir_path, 0o777)

            return False

        try:
            file = open(self.file_path, 'r')
            self._xml.setContent(file.read())
            file.close()
            return True

        except:
            self._remove_file()
            return False

    def _write_file(self):
        try:
            file = open(self.file_path, 'w')
            file.write(self._xml.toString())
            file.close()
        except:
            return

    def _set_attributes(self, element: QDomElement):
        element.setAttribute('net_daemon_id', self.server.net_daemon_id)
        element.setAttribute('root', self.session.root)
        element.setAttribute('session_path', self.session.path)
        element.setAttribute('pid', os.getpid())
        element.setAttribute('port', self.server.port)
        element.setAttribute('user', os.getenv('USER'))
        element.setAttribute('not_default',
                             int(bool(self.server.is_nsm_locked
                                      or self.server.not_default)))
        element.setAttribute('has_gui', int(self.server.has_gui()))
        element.setAttribute('version', ray.VERSION)
        element.setAttribute('local_gui_pids',
                             self.server.get_local_gui_pid_list())

        for locked_path in self._locked_session_paths:
            locked_paths_xml = self._xml.createElement('locked_session')
            locked_paths_xml.setAttribute('path', locked_path)
            element.appendChild(locked_paths_xml)

    def _clean_dirty_pids(self):
        xml_content = self._xml.documentElement()
        nodes = xml_content.childNodes()
        rm_nodes = []

        for i in range(nodes.count()):
            node = nodes.at(i)
            dxe = node.toElement()
            pid = dxe.attribute('pid')
            if not pid.isdigit() or not self._pid_exists(int(pid)):
                rm_nodes.append(node)

        for node in rm_nodes:
            xml_content.removeChild(node)

    def update(self):
        has_dirty_pid = False

        if not self._open_file():
            ds = self._xml.createElement('Daemons')
            dm_xml = self._xml.createElement('Daemon')

            self._set_attributes(dm_xml)

            ds.appendChild(dm_xml)
            self._xml.appendChild(ds)

        else:
            xml_content = self._xml.documentElement()
            nodes = xml_content.childNodes()
            self_node = None

            for i in range(nodes.count()):
                node = nodes.at(i)
                dxe = node.toElement()
                pid = dxe.attribute('pid')

                if pid.isdigit() and pid == str(os.getpid()):
                    self_node = node
                elif not pid.isdigit() or not self._pid_exists(int(pid)):
                    has_dirty_pid = True
            
            if self_node is not None:
                xml_content.removeChild(self_node)

            dm_xml = self._xml.createElement('Daemon')
            self._set_attributes(dm_xml)
            self._xml.firstChild().appendChild(dm_xml)

        if has_dirty_pid:
            self._clean_dirty_pids()

        self._write_file()

    def quit(self):
        if not self._open_file():
            return

        xml_content = self._xml.documentElement()
        nodes = xml_content.childNodes()

        for i in range(nodes.count()):
            node = nodes.at(i)
            dxe = node.toElement()
            pid = dxe.attribute('pid')

            if pid.isdigit() and pid == str(os.getpid()):
                break
        else:
            return

        xml_content.removeChild(node)
        self._write_file()

    def is_free_for_root(self, daemon_id, root_path)->bool:
        if not self._open_file():
            return True

        xml_content = self._xml.documentElement()
        nodes = xml_content.childNodes()

        for i in range(nodes.count()):
            node = nodes.at(i)
            dxe = node.toElement()
            if (dxe.attribute('net_daemon_id') == str(daemon_id)
                    and dxe.attribute('root') == root_path):
                pid = dxe.attribute('pid')
                if pid.isdigit() and self._pid_exists(int(pid)):
                    return False

        return True

    def is_free_for_session(self, session_path: Union[str, Path]) -> bool:
        session_path = str(session_path)
        
        if not self._open_file():
            return True

        xml_content = self._xml.documentElement()
        nodes = xml_content.childNodes()

        for i in range(nodes.count()):
            node = nodes.at(i)
            dxe = node.toElement()
            pid = dxe.attribute('pid')

            if dxe.attribute('session_path') == session_path:                
                if pid.isdigit() and self._pid_exists(int(pid)):
                    return False
            
            sub_nodes = node.childNodes()
            for j in range(sub_nodes.count()):
                sub_node = sub_nodes.at(j)
                sub_nod_el = sub_node.toElement()
                if sub_nod_el.attribute('path') == session_path:
                    if pid.isdigit() and self._pid_exists(int(pid)):
                        return False

        return True

    def get_all_session_paths(self)->list:
        if not self._open_file():
            return []

        all_session_paths = []

        xml_content = self._xml.documentElement()
        nodes = xml_content.childNodes()

        for i in range(nodes.count()):
            node = nodes.at(i)
            dxe = node.toElement()
            spath = dxe.attribute('session_path')
            pid = dxe.attribute('pid')
            if spath and pid.isdigit() and self._pid_exists(int(pid)):
                all_session_paths.append(spath)

        return all_session_paths

    def add_locked_path(self, path:str):
        self._locked_session_paths.add(path)
        self.update()
    
    def unlock_path(self, path:str):
        self._locked_session_paths.discard(path)
        self.update()

    def get_daemon_list(self)->list:
        daemon_list = []
        has_dirty_pid = False

        if not self._open_file():
            return daemon_list

        xml_content = self._xml.documentElement()
        nodes = xml_content.childNodes()

        for i in range(nodes.count()):
            node = nodes.at(i)
            dxe = node.toElement()

            daemon = Daemon()
            daemon.root = dxe.attribute('root')
            daemon.session_path = dxe.attribute('session_path')
            daemon.user = dxe.attribute('user')
            daemon.not_default = bool(dxe.attribute('not_default') == 'true')
            net_daemon_id = dxe.attribute('net_daemon_id')
            pid = dxe.attribute('pid')
            port = dxe.attribute('port')

            if net_daemon_id.isdigit():
                daemon.net_daemon_id = net_daemon_id
            if pid.isdigit():
                daemon.pid = pid
            if port.isdigit():
                daemon.port = port

            if not self._pid_exists(daemon.pid):
                has_dirty_pid = True
                continue

            if not (daemon.net_daemon_id
                    and daemon.pid
                    and daemon.port):
                continue

            daemon_list.append(daemon)

        if has_dirty_pid:
            self._clean_dirty_pids()

        return daemon_list




from typing import TYPE_CHECKING
from PyQt5.QtGui import QIcon, QDesktopServices
from PyQt5.QtWidgets import QMenu, QApplication
from PyQt5.QtCore import QLocale, QUrl

from . import patchcanvas
from .base_elements import PortType, PortTypesViewFlag

if TYPE_CHECKING:
    from .patchbay_manager import PatchbayManager

_translate = QApplication.translate


class CanvasMenu(QMenu):
    def __init__(self, patchbay_manager: 'PatchbayManager'):
        QMenu.__init__(self, _translate('patchbay', 'Patchbay'))
        self.patchbay_manager = patchbay_manager

        # fix wrong menu position with Wayland,
        # see https://community.kde.org/Guidelines_and_HOWTOs/Wayland_Porting_Notes
        self.winId()
        main_win = self.patchbay_manager.main_win
        main_win.winId()
        parent_window_handle = main_win.windowHandle()
        if not parent_window_handle:
            native_parent_widget = main_win.nativeParentWidget()
            if native_parent_widget:
                parent_window_handle = native_parent_widget.windowHandle()
        self.windowHandle().setTransientParent(parent_window_handle)

        self.patchbay_manager.sg.port_types_view_changed.connect(
            self._port_types_view_changed)
        self.patchbay_manager.sg.alsa_midi_enabled_changed.connect(
            self._alsa_midi_enabled)

        self.action_fullscreen = self.addAction(
            _translate('patchbay', "Toggle Full Screen"))
        self.action_fullscreen.setIcon(QIcon.fromTheme('view-fullscreen'))
        self.action_fullscreen.triggered.connect(
            patchbay_manager.sg.full_screen_toggle_wanted.emit)

        port_types_view = patchbay_manager.port_types_view & (
            PortType.AUDIO_JACK | PortType.MIDI_JACK)

        self.action_find_box = self.addAction(
            _translate('patchbay', "Find a box...\tCtrl+F"))
        self.action_find_box.setIcon(QIcon.fromTheme('edit-find'))
        self.action_find_box.triggered.connect(
            patchbay_manager.sg.filters_bar_toggle_wanted.emit)

        self.port_types_menu = QMenu(_translate('patchbay', 'Type filter'), self)
        self.port_types_menu.setIcon(QIcon.fromTheme('view-filter'))
        self.action_all_types = self.port_types_menu.addAction(
            _translate('patchbay', 'AUDIO | MIDI | CV'))
        self.action_all_types.setCheckable(True)
        self.action_all_types.setChecked(
            bool(port_types_view is PortTypesViewFlag.ALL))
        self.action_all_types.triggered.connect(
            self.port_types_view_all_types_choice)

        self.action_audio = self.port_types_menu.addAction(
            _translate('patchbay', 'AUDIO only'))
        self.action_audio.setCheckable(True)
        self.action_audio.setChecked(port_types_view is PortTypesViewFlag.AUDIO)
        self.action_audio.triggered.connect(
            self.port_types_view_audio_choice)

        self.action_midi = self.port_types_menu.addAction(
            _translate('patchbay', 'MIDI only'))
        self.action_midi.setCheckable(True)
        self.action_midi.setChecked(port_types_view is PortTypesViewFlag.MIDI)
        self.action_midi.triggered.connect(
            self.port_types_view_midi_choice)
        
        self.action_cv = self.port_types_menu.addAction(
            _translate('patchbay', 'CV only'))
        self.action_cv.setCheckable(True)
        self.action_cv.setChecked(port_types_view is PortTypesViewFlag.CV)
        self.action_cv.triggered.connect(
            self.port_types_view_cv_choice)

        self.action_alsa = self.port_types_menu.addAction(
            _translate('patchbay', 'ALSA only'))
        self.action_alsa.setCheckable(True)
        self.action_alsa.setChecked(port_types_view is PortTypesViewFlag.ALSA)
        self.action_alsa.triggered.connect(
            self.port_types_view_alsa_choice)

        self.addMenu(self.port_types_menu)

        self._alsa_midi_enabled(patchbay_manager.alsa_midi_enabled)

        self.zoom_menu = QMenu(_translate('patchbay', 'Zoom'), self)
        self.zoom_menu.setIcon(QIcon.fromTheme('zoom'))

        self.autofit = self.zoom_menu.addAction(
            _translate('patchbay', 'auto-fit'))
        self.autofit.setIcon(QIcon.fromTheme('zoom-select-fit'))
        self.autofit.setShortcut('Home')
        self.autofit.triggered.connect(patchcanvas.canvas.scene.zoom_fit)

        self.zoom_in = self.zoom_menu.addAction(
            _translate('patchbay', 'Zoom +'))
        self.zoom_in.setIcon(QIcon.fromTheme('zoom-in'))
        self.zoom_in.setShortcut('Ctrl++')
        self.zoom_in.triggered.connect(patchcanvas.canvas.scene.zoom_in)

        self.zoom_out = self.zoom_menu.addAction(
            _translate('patchbay', 'Zoom -'))
        self.zoom_out.setIcon(QIcon.fromTheme('zoom-out'))
        self.zoom_out.setShortcut('Ctrl+-')
        self.zoom_out.triggered.connect(patchcanvas.canvas.scene.zoom_out)

        self.zoom_orig = self.zoom_menu.addAction(
            _translate('patchbay', 'Default Zoom'))
        self.zoom_orig.setIcon(QIcon.fromTheme('zoom'))
        self.zoom_orig.setShortcut('Ctrl+1')
        self.zoom_orig.triggered.connect(patchcanvas.canvas.scene.zoom_reset)

        self.addMenu(self.zoom_menu)

        self.action_refresh = self.addAction(
            _translate('patchbay', "Refresh the canvas\tCtrl+R"))
        self.action_refresh.setIcon(QIcon.fromTheme('view-refresh'))
        self.action_refresh.triggered.connect(patchbay_manager.refresh)

        self.action_manual = self.addAction(
            _translate('patchbay', "Patchbay manual"))
        self.action_manual.setIcon(QIcon.fromTheme('system-help'))
        self.action_manual.triggered.connect(self.internal_manual)
        self.action_manual.setVisible(
            self.patchbay_manager._manual_path is not None)

        self.action_options = self.addAction(
            _translate('patchbay', "Canvas options"))
        self.action_options.setIcon(QIcon.fromTheme("configure"))
        self.action_options.triggered.connect(
            patchbay_manager.show_options_dialog)

    def _port_types_view_changed(self, port_types_view: int):
        self.action_all_types.setChecked(
            port_types_view == PortTypesViewFlag.ALL)
        self.action_audio.setChecked(
            port_types_view == PortTypesViewFlag.AUDIO)
        self.action_midi.setChecked(
            port_types_view == PortTypesViewFlag.MIDI)
        self.action_cv.setChecked(
            port_types_view == PortTypesViewFlag.CV)
        self.action_alsa.setChecked(
            port_types_view == PortTypesViewFlag.ALSA)

    def port_types_view_all_types_choice(self):
        self.patchbay_manager.change_port_types_view(
            PortTypesViewFlag.ALL)

    def port_types_view_audio_choice(self):
        self.patchbay_manager.change_port_types_view(
            PortTypesViewFlag.AUDIO)

    def port_types_view_midi_choice(self):
        self.patchbay_manager.change_port_types_view(
            PortTypesViewFlag.MIDI)

    def port_types_view_cv_choice(self):
        self.patchbay_manager.change_port_types_view(
            PortTypesViewFlag.CV)

    def port_types_view_alsa_choice(self):
        self.patchbay_manager.change_port_types_view(
            PortTypesViewFlag.ALSA)

    def _alsa_midi_enabled(self, yesno: bool):
        self.action_alsa.setVisible(yesno)
        if yesno:
            self.action_all_types.setText(
                _translate('patchbay', 'AUDIO | MIDI | CV | ALSA'))
        else:
            self.action_all_types.setText(
                _translate('patchbay', 'AUDIO | MIDI | CV'))

    def internal_manual(self):
        short_locale = 'en'
        manual_dir = self.patchbay_manager._manual_path
        if manual_dir is None:
            return        
        
        locale_str = QLocale.system().name()
        html_path = manual_dir / locale_str[:2] / 'manual.html'
        
        if (len(locale_str) > 2 and '_' in locale_str
                and html_path.is_file()):
            short_locale = locale_str[:2]

        url = QUrl(f"file://{manual_dir}/{short_locale}/manual.html")
        QDesktopServices.openUrl(url)



import logging
import os
import shlex
import sys
import random
import shutil
import subprocess
import time
from typing import TYPE_CHECKING
from pathlib import Path
import xml.etree.ElementTree as ET

try:
    import liblo
except ImportError:
    import pyliblo3 as liblo
from PyQt5.QtCore import QCoreApplication

import ray
from signaler import Signaler
from multi_daemon_file import MultiDaemonFile
from daemon_tools import (
    TemplateRoots,
    CommandLineArgs,
    Terminal,
    RS,
    get_code_root)
from xml_tools import XmlElement
from terminal_starter import which_terminal

if TYPE_CHECKING:
    from session_signaled import SignaledSession

instance = None
signaler = Signaler.instance()
_translate = QCoreApplication.translate
_logger = logging.getLogger(__name__)

def _path_is_valid(path: str) -> bool:
    if path.startswith(('./', '../')):
        return False

    for forbidden in ('//', '/./', '/../'):
        if forbidden in path:
            return False

    if path.endswith(('/.', '/..')):
        return False
    return True

def ray_method(path, types):
    def decorated(func):
        @liblo.make_method(path, types)
        def wrapper(*args, **kwargs):
            t_thread, t_path, t_args, t_types, src_addr, rest = args
            if CommandLineArgs.debug:
                sys.stderr.write(
                    '\033[94mOSC::daemon_receives\033[0m %s, %s, %s, %s\n'
                    % (t_path, t_types, t_args, src_addr.url))

            response = func(*args[:-1], **kwargs)
            if response != False:
                signaler.osc_recv.emit(t_path, t_args, t_types, src_addr)

            return response
        return wrapper
    return decorated


class Controller:
    addr: liblo.Address = None
    pid = 0


class GuiAdress(liblo.Address):
    gui_pid = 0


# Osc server thread separated in many classes for confort.

# ClientCommunicating contains NSM protocol.
# OSC paths have to be never changed.
class ClientCommunicating(liblo.ServerThread):
    def __init__(self, session: 'SignaledSession', osc_num=0):
        liblo.ServerThread.__init__(self, osc_num)
        self.session = session

        self._nsm_locker_url = ''
        self._net_master_daemon_url = ''
        self._list_asker_addr = None

        self.gui_list = list[GuiAdress]()
        self.controller_list = list[Controller]()
        self.monitor_list = list[liblo.Address]()
        self.server_status = ray.ServerStatus.OFF
        self.is_nsm_locked = False
        self.not_default = False

        self.net_daemon_id = random.randint(1, 999999999)
        self.options = 0

    @ray_method('/osc/ping', '')
    def oscPing(self, path, args, types, src_addr):
        self.send(src_addr, "/reply", path)

    @ray_method('/reply', None)
    def reply(self, path, args, types, src_addr):
        if not ray.types_are_all_strings(types):
            self._unknown_message(path, types, src_addr)

        if not len(args) >= 1:
            self._unknown_message(path, types, src_addr)
            return False

        reply_path = args[0]

        if reply_path == '/ray/server/list_sessions':
            # this reply is only used here for reply from net_daemon
            # it directly resend its infos
            # to the last addr that asked session list
            if self._list_asker_addr:
                self.send(self._list_asker_addr, path, *args)
            return False

        if reply_path == '/ray/gui/script_user_action':
            self.send_gui('/ray/gui/hide_script_user_action')
            for controller in self.controller_list:
                self.send(controller.addr, '/reply',
                          '/ray/server/script_user_action',
                          'User action dialog validate')
            return False

        if not len(args) == 2:
            # assume this is a normal client, not a net_daemon
            self._unknown_message(path, types, src_addr)
            return False

    @ray_method('/error', 'sis')
    def error(self, path, args, types, src_addr):
        error_path, error_code, error_string = args

        if error_path == '/ray/gui/script_user_action':
            self.send_gui('/ray/gui/hide_script_user_action')

            for controller in self.controller_list:
                self.send(controller.addr, '/error',
                          '/ray/server/script_user_action', -1,
                          'User action dialog aborted !')
            return False

    @ray_method('/minor_error', 'sis')
    def minor_error(self, path, args, types, src_addr):
        # prevent minor_error to minor_error loop in daemon <-> daemon communication
        pass

    # SERVER_CONTROL messages
    # following messages only for :server-control: capability
    @ray_method('/nsm/server/add', 's')
    def nsmServerAdd(self, path, args, types, src_addr):
        executable_path = args[0]

        if not self.session.path:
            self.send(src_addr, "/error", path, ray.Err.NO_SESSION_OPEN,
                      "Cannot add to session because no session is loaded.")
            return False

        if '/' in executable_path:
            self.send(src_addr, "/error", path, ray.Err.LAUNCH_FAILED,
                "Absolute paths are not permitted. Clients must be in $PATH")
            return False

    @ray_method('/nsm/server/save', '')
    def nsmServerSave(self, path, args, types, src_addr):
        if not self.session.path:
            self.send(src_addr, "/error", path, ray.Err.NO_SESSION_OPEN,
                      "No session to save.")
            return False

    @ray_method('/nsm/server/open', 's')
    def nsmServerOpen(self, path, args, types, src_addr):
        pass

    @ray_method('/nsm/server/new', 's')
    def nsmServerNew(self, path, args, types, src_addr):
        if self.is_nsm_locked:
            return False

        if not _path_is_valid(args[0]):
            self.send(src_addr, "/error", path, ray.Err.CREATE_FAILED,
                      "Invalid session name.")
            return False

    @ray_method('/nsm/server/duplicate', 's')
    def nsmServerDuplicate(self, path, args, types, src_addr):
        if self.is_nsm_locked or not self.session.path:
            self.send(src_addr, "/error", path, ray.Err.NO_SESSION_OPEN,
                      "No session to duplicate.")
            return False

        if not _path_is_valid(args[0]):
            self.send(src_addr, "/error", path, ray.Err.CREATE_FAILED,
                      "Invalid session name.")
            return False

    @ray_method('/nsm/server/close', '')
    def nsmServerClose(self, path, args, types, src_addr):
        if not self.session.path:
            self.send(src_addr, "/error", path, ray.Err.NO_SESSION_OPEN,
                      "No session to close.")
            return False

    @ray_method('/nsm/server/abort', '')
    def nsmServerAbort(self, path, args, types, src_addr):
        if not self.session.path:
            self.send(src_addr, "/error", path, ray.Err.NO_SESSION_OPEN,
                      "No session to abort.")
            return False

    @ray_method('/nsm/server/quit', '')
    def nsmServerQuit(self, path, args, types, src_addr):
        pass

    @ray_method('/nsm/server/list', '')
    def nsmServerList(self, path, args, types, src_addr):
        pass
    # END OF SERVER_CONTROL messages

    @ray_method('/nsm/server/announce', 'sssiii')
    def nsmServerAnnounce(self, path, args, types, src_addr):
        if not self.session.path:
            self.send(src_addr, "/error", path, ray.Err.NO_SESSION_OPEN,
                      "Sorry, but there's no session open "
                      + "for this application to join.")
            return False

    @ray_method('/nsm/server/broadcast', None)
    def nsmServerBroadcast(self, path, args, types, src_addr: liblo.Address):
        if not args:
            return False

        if not isinstance(args[0], str):
            return False

        #don't allow clients to broadcast NSM commands
        follow_path = args[0]
        if not isinstance(follow_path, str):
            return False
        if follow_path.startswith(('/nsm/', '/ray/')):
            return False

        for client in self.session.clients:
            if not client.addr:
                continue

            if not ray.are_same_osc_port(client.addr.url, src_addr.url):
                self.send(client.addr, liblo.Message(*args))

            # TODO broadcast to slave daemons
            #for gui_addr in self.gui_list:
                ##also relay to attached GUI so that the broadcast can be
                ##propagated to another NSMD instance
                #if gui_addr.url != src_addr.url:
                    #self.send(gui_addr, Message(*args))

    @ray_method('/nsm/server/monitor_reset', '')
    def nsmServerGetAllStates(self, path, args, types, src_addr):
        self.send(src_addr, '/reply', path, 'monitor reset')
        self.session.send_initial_monitor(src_addr, monitor_is_client=True)

    @ray_method('/nsm/client/progress', 'f')
    def nsmClientProgress(self, path, args, types, src_addr):
        client = self.session.get_client_by_address(src_addr)
        if not client:
            return False

        client.progress = args[0]
        self.send_gui("/ray/gui/client/progress", client.client_id,
                     client.progress)

    @ray_method('/nsm/client/is_dirty', '')
    def nsmClientIs_dirty(self, path, args, types, src_addr):
        client = self.session.get_client_by_address(src_addr)
        if not client:
            return False

        Terminal.message("%s sends dirty" % client.client_id)

        client.dirty = 1
        client.last_dirty = time.time()

        self.send_gui("/ray/gui/client/dirty", client.client_id, client.dirty)

    @ray_method('/nsm/client/is_clean', '')
    def nsmClientIs_clean(self, path, args, types, src_addr):
        client = self.session.get_client_by_address(src_addr)
        if not client:
            return False

        Terminal.message("%s sends clean" % client.client_id)

        client.dirty = 0

        self.send_gui("/ray/gui/client/dirty", client.client_id, client.dirty)
        return False

    @ray_method('/nsm/client/message', 'is')
    def nsmClientMessage(self, path, args, types, src_addr):
        client = self.session.get_client_by_address(src_addr)
        if not client:
            return False

        self.send_gui("/ray/gui/client/message",
                      client.client_id, args[0], args[1])

    @ray_method('/nsm/client/gui_is_hidden', '')
    def nsmClientGui_is_hidden(self, path, args, types, src_addr):
        client = self.session.get_client_by_address(src_addr)
        if not client:
            return False

        Terminal.message("Client '%s' sends gui hidden" % client.client_id)

        client.gui_visible = False

        self.send_gui("/ray/gui/client/gui_visible", client.client_id,
                     int(client.gui_visible))

    @ray_method('/nsm/client/gui_is_shown', '')
    def nsmClientGui_is_shown(self, path, args, types, src_addr):
        client = self.session.get_client_by_address(src_addr)
        if not client:
            return False

        Terminal.message("Client '%s' sends gui shown" % client.client_id)

        client.gui_visible = True
        client.gui_has_been_visible = True

        self.send_gui("/ray/gui/client/gui_visible", client.client_id,
                     int(client.gui_visible))

    @ray_method('/nsm/client/label', 's')
    def nsmClientLabel(self, path, args, types, src_addr):
        pass

    @ray_method('/nsm/client/network_properties', 'ss')
    def nsmClientNetworkProperties(self, path, args, types, src_addr):
        pass

    @ray_method('/nsm/client/no_save_level', 'i')
    def nsmClientNoSaveLevel(self, path, args, types, src_addr):
        pass

    def _unknown_message(self, path, types, src_addr):
        self.send(src_addr, '/minor_error', path, ray.Err.UNKNOWN_MESSAGE,
                  "unknown osc message: %s %s" % (path, types))

    def send_gui(self, *args):
        # should be overclassed
        pass

class OscServerThread(ClientCommunicating):
    def __init__(self, session, osc_num=0):
        ClientCommunicating.__init__(self, session, osc_num)

        self.options_dict = {
            'save_from_client': ray.Option.SAVE_FROM_CLIENT,
            'bookmark_session_folder': ray.Option.BOOKMARK_SESSION,
            'desktops_memory': ray.Option.DESKTOPS_MEMORY,
            'snapshots': ray.Option.SNAPSHOTS,
            'session_scripts': ray.Option.SESSION_SCRIPTS,
            'gui_states': ray.Option.GUI_STATES}

        self.options = RS.settings.value(
            'daemon/options',
            ray.Option.BOOKMARK_SESSION
            + ray.Option.SNAPSHOTS
            + ray.Option.SESSION_SCRIPTS,
            type=int
        )

        if CommandLineArgs.no_options:
            self.options = 0

        if shutil.which('wmctrl'):
            self.options |= ray.Option.HAS_WMCTRL
        else:
            self.options &= ~ray.Option.HAS_WMCTRL

        if shutil.which('git'):
            self.options |= ray.Option.HAS_GIT
        else:
            self.options &= ~ray.Option.HAS_GIT

        self.client_templates_database = {
            'factory': [], 'user': []}

        self.session_to_preview = ''
        
        self._terminal_command_is_default = True
        self.terminal_command = RS.settings.value(
            'daemon/terminal_command', '', type=str)
        if self.terminal_command:
            self._terminal_command_is_default = False
        else:
            self.terminal_command = shlex.join(
                which_terminal(title='RAY_TERMINAL_TITLE'))

        global instance
        instance = self

    @staticmethod
    def get_instance() -> 'OscServerThread':
        return instance

    @ray_method('/ray/server/gui_announce', 'sisii')
    def rayGuiGui_announce(self, path, args, types, src_addr: liblo.Address):
        (version, int_nsm_locked, net_master_daemon_url,
         gui_pid, net_daemon_id) = args

        nsm_locked = bool(int_nsm_locked)
        is_net_free = True

        if nsm_locked:
            self._net_master_daemon_url = net_master_daemon_url
            self.is_nsm_locked = True
            self._nsm_locker_url = src_addr.url

            for gui_addr in self.gui_list:
                if not ray.are_same_osc_port(gui_addr.url, src_addr.url):
                    self.send(gui_addr, '/ray/gui/server/nsm_locked', 1)

            self.net_daemon_id = net_daemon_id

            multi_daemon_file = MultiDaemonFile.get_instance()
            if multi_daemon_file:
                is_net_free = multi_daemon_file.is_free_for_root(
                    self.net_daemon_id, self.session.root)

        self.announce_gui(src_addr.url, nsm_locked, is_net_free, gui_pid)

    @ray_method('/ray/server/gui_disannounce', '')
    def rayGuiGui_disannounce(self, path, args, types, src_addr: liblo.Address):
        for addr in self.gui_list:
            if ray.are_same_osc_port(addr.url, src_addr.url):
                break
        else:
            return False

        self.gui_list.remove(addr)

        if src_addr.url == self._nsm_locker_url:
            self.net_daemon_id = random.randint(1, 999999999)

            self.is_nsm_locked = False
            self._nsm_locker_url = ''
            self.send_gui('/ray/gui/server/nsm_locked', 0)

        multi_daemon_file = MultiDaemonFile.get_instance()
        if multi_daemon_file:
            multi_daemon_file.update()

    @ray_method('/ray/server/ask_for_patchbay', '')
    def rayServerGetPatchbayPort(self, path, args, types, src_addr):
        patchbay_file = '/tmp/RaySession/patchbay_daemons/' + str(self.port)

        if not os.path.exists(patchbay_file):
            return True

        with open(patchbay_file, 'r') as file:
            file = open(patchbay_file, 'r')
            contents = file.read()
            #file.close()
            for line in contents.splitlines():
                if line.startswith('pid:'):
                    pid_str = line.rpartition(':')[2]
                    if pid_str.isdigit():
                        pid = int(pid_str)
                        try:
                            os.kill(pid, 0)
                        except OSError:
                            # go to main thread (session_signaled.py)
                            return True
                        else:
                            # pid is okay, let check the osc port next
                            continue
                    else:
                        return True

                if line.startswith('port:'):
                    port_str = line.rpartition(':')[2]
                    good_port = False

                    try:
                        patchbay_addr = liblo.Address(int(port_str))
                        good_port = True
                    except:
                        patchbay_addr = None
                        sys.stderr.write(
                            'port given for patchbay %s is not a valid osc port')

                    if good_port:
                        self.send(patchbay_addr, '/ray/patchbay/add_gui',
                                  src_addr.url)
                        return False
                    break

        # continue in main thread if patchbay_to_osc is not started yet
        # see session_signaled.py -> _ray_server_ask_for_patchbay

    @ray_method('/ray/server/controller_announce', 'i')
    def rayServerControllerAnnounce(self, path, args, types, src_addr):
        controller = Controller()
        controller.addr = src_addr
        controller.pid = args[0]
        self.controller_list.append(controller)
        self.send(src_addr, '/reply', path, 'announced')

    @ray_method('/ray/server/controller_disannounce', '')
    def rayServerControllerDisannounce(self, path, args, types, src_addr: liblo.Address):
        for controller in self.controller_list:
            if controller.addr.url == src_addr.url:
                break
        else:
            return

        self.controller_list.remove(controller)
        self.send(src_addr, '/reply', path, 'disannounced')

    @ray_method('/ray/server/monitor_announce', '')
    def rayServerMonitorAnnounce(self, path, args, types, src_addr):
        monitor_addr = src_addr
        self.monitor_list.append(monitor_addr)
        self.session.send_initial_monitor(src_addr, monitor_is_client=False)
        self.send(src_addr, '/reply', path, 'announced')
    
    @ray_method('/ray/server/monitor_quit', '')
    def rayServerMonitorDisannounce(self, path, args, types, src_addr: liblo.Address):
        for monitor_addr in self.monitor_list:
            if monitor_addr.url == src_addr.url:
                break
        else:
            return
        
        self.monitor_list.remove(monitor_addr)
        self.send(src_addr, '/reply', path, 'monitor exit')

    @ray_method('/ray/server/set_nsm_locked', '')
    def rayServerSetNsmLocked(self, path, args, types, src_addr: liblo.Address):
        self.is_nsm_locked = True
        self._nsm_locker_url = src_addr.url

        for gui_addr in self.gui_list:
            if gui_addr.url != src_addr.url:
                self.send(gui_addr, '/ray/gui/server/nsm_locked', 1)

    @ray_method('/ray/server/quit', '')
    def rayServerQuit(self, path, args, types, src_addr):
        pass

    @ray_method('/ray/server/abort_copy', '')
    def rayServerAbortCopy(self, path, args, types, src_addr):
        pass

    @ray_method('/ray/server/abort_parrallel_copy', 'i')
    def rayServerAbortParrallelCopy(self, path, args, types, src_addr):
        pass

    @ray_method('/ray/server/abort_snapshot', '')
    def rayServerAbortSnapshot(self, path, args, types, src_addr):
        pass

    @ray_method('/ray/server/change_root', 's')
    def rayServerChangeRoot(self, path, args, types, src_addr):
        new_root: str = args[0]
        if not(new_root.startswith('/') and _path_is_valid(new_root)):
            self.send(src_addr, '/error', path, ray.Err.CREATE_FAILED,
                      "invalid session root !")

        if self._is_operation_pending(src_addr, path):
            self.send(src_addr, '/error', path, ray.Err.OPERATION_PENDING,
                      "Can't change session_root. Operation pending")
            return False

    @ray_method('/ray/server/set_terminal_command', 's')
    def rayServerSetTerminalCommand(self, path, args, types, src_addr):
        if args[0] != self.terminal_command:
            self.terminal_command = args[0]
            if not self.terminal_command:
                self.terminal_command = shlex.join(
                    which_terminal(title='RAY_TERMINAL_TITLE'))
            self.send_gui('/ray/gui/server/terminal_command',
                          self.terminal_command)
        self.send(src_addr, '/reply', path, 'terminal command set')

    @ray_method('/ray/server/list_path', '')
    def rayServerListPath(self, path, args, types, src_addr):
        exec_list = list[str]()
        tmp_exec_list = list[str]()
        n = 0

        pathlist = os.getenv('PATH').split(':')
        for pathdir in pathlist:
            if os.path.isdir(pathdir):
                listexe = os.listdir(pathdir)
                for exe in listexe:
                    fullexe = pathdir + '/' + exe

                    if (os.path.isfile(fullexe)
                            and os.access(fullexe, os.X_OK)
                            and not exe in exec_list):
                        exec_list.append(exe)
                        tmp_exec_list.append(exe)
                        n += len(exe)

                        if n >= 20000:
                            self.send(src_addr, '/reply',
                                      path, *tmp_exec_list)
                            tmp_exec_list.clear()
                            n = 0

        if tmp_exec_list:
            self.send(src_addr, '/reply', path, *tmp_exec_list)

    @ray_method('/ray/server/list_session_templates', '')
    def rayServerListSessionTemplates(self, path, args, types, src_addr):
        if not TemplateRoots.user_sessions.is_dir():
            self.send(src_addr, '/reply', path)
            return False

        template_list = list[str]()

        all_files = os.listdir(TemplateRoots.user_sessions)
        for file in all_files:
            if os.path.isdir("%s/%s" % (TemplateRoots.user_sessions, file)):
                template_list.append(file)

                if len(template_list) == 100:
                    self.send(src_addr, '/reply', path, *template_list)
                    template_list.clear()

        if template_list:
            self.send(src_addr, '/reply', path, *template_list)

        self.send(src_addr, '/reply', path)

    @ray_method('/ray/server/list_user_client_templates', None)
    def rayServerListUserClientTemplates(self, path, args, types, src_addr):
        for a in types:
            if a != 's':
                self._unknown_message(path, types, src_addr)
                return False

    @ray_method('/ray/server/list_factory_client_templates', None)
    def rayServerListFactoryClientTemplates(self, path, args, types,
                                            src_addr):
        for a in types:
            if a != 's':
                self._unknown_message(path, types, src_addr)
                return False

    @ray_method('/ray/server/remove_client_template', 's')
    def rayServerRemoveClientTemplate(self, path, args, types, src_addr):
        template_name: str = args[0]
        templates_root = TemplateRoots.user_clients
        templates_file = templates_root / 'client_templates.xml'

        if not templates_file.is_file():
            self.send(src_addr, '/error', path, ray.Err.NO_SUCH_FILE,
                      "file %s is missing !" % templates_file)
            return False

        if not os.access(templates_file, os.W_OK):
            self.send(src_addr, '/error', path, ray.Err.NO_SUCH_FILE,
                      "file %s in unwriteable !" % templates_file)
            return False

        tree = ET.parse(templates_file)
        root = tree.getroot()

        if root.tag != 'RAY-CLIENT-TEMPLATES':
            self.send(src_addr, '/error', path, ray.Err.BAD_PROJECT,
                      "file %s is not write correctly !" % templates_file)
            return False

        xroot = XmlElement(root)

        for c in xroot.iter():
            if c.el.tag != 'Client-Template':
                continue
            
            if c.str('template-name') == template_name:
                break
        else:
            self.send(src_addr, '/error', path, ray.Err.NO_SUCH_FILE,
                      "No template \"%s\" to remove !" % template_name)
            return False
        
        root.remove(c.el)
        
        try:
            tree.write(templates_file)
        except BaseException as e:
            _logger.error(str(e))
            self.send(src_addr, '/error', path, ray.Err.CREATE_FAILED,
                      "Impossible to rewrite user client templates xml file")
            return False
        
        templates_dir = templates_root / template_name
        if templates_dir.is_dir():
            try:
                shutil.rmtree(templates_dir)
            except BaseException as e:
                _logger.error(str(e))
                self.send(src_addr, '/error', path, ray.Err.CREATE_FAILED,
                      "Failed to remove the folder %s" % str(templates_dir))
                return False

        self.send(src_addr, '/reply', path,
                  f'template "{template_name}" removed.')

    @ray_method('/ray/server/list_sessions', '')
    def rayServerListSessions(self, path, args, types, src_addr):
        self._list_asker_addr = src_addr

    @ray_method('/ray/server/list_sessions', 'i')
    def rayServerListSessionsWithNet(self, path, args, types, src_addr):
        self._list_asker_addr = src_addr

    @ray_method('/ray/server/new_session', None)
    def rayServerNewSession(self, path, args, types, src_addr):
        if not ray.types_are_all_strings(types):
            self._unknown_message(path, types, src_addr)
            return False

        if self.is_nsm_locked:
            return False

        if not _path_is_valid(args[0]):
            self.send(src_addr, "/error", path, ray.Err.CREATE_FAILED,
                      "Invalid session name.")
            return False

    @ray_method('/ray/server/open_session', 's')
    def rayServerOpenSession(self, path, args, types, src_addr):
        pass

    @ray_method('/ray/server/open_session', 'si')
    def rayServerOpenSessionWithoutSave(self, path, args, types, src_addr):
        pass

    @ray_method('/ray/server/open_session', 'sis')
    def rayServerOpenSessionWithTemplate(self, path, args, types, src_addr):
        pass

    @ray_method('/ray/server/open_session_off', 's')
    def rayServerOpenSessionOff(self, path, args, types, src_addr):
        pass

    @ray_method('/ray/server/open_session_off', 'si')
    def rayServerOpenSessionWithoutSaveOff(self, path, args, types, src_addr):
        pass

    @ray_method('/ray/server/save_session_template', 'ss')
    def rayServerSaveSessionTemplate(self, path, args, types, src_addr):
        #save as template an not loaded session
        session_name, template_name = args

        if not _path_is_valid(session_name):
            self.send(src_addr, "/error", path, ray.Err.CREATE_FAILED,
                    "Invalid session name.")
            return False

        if '/' in template_name:
            self.send(src_addr, "/error", path, ray.Err.CREATE_FAILED,
                      "Invalid template name.")
            return False

    @ray_method('/ray/server/rename_session', 'ss')
    def rayServerRenameSession(self, path, args, types, src_addr):
        pass

    @ray_method('/ray/server/save_session_template', 'sss')
    def rayServerSaveSessionTemplateWithRoot(self, path, args,
                                             types, src_addr):
        #save as template an not loaded session
        session_name, template_name, sess_root = args
        if not _path_is_valid(session_name):
            self.send(src_addr, "/error", path, ray.Err.CREATE_FAILED,
                    "Invalid template name.")
            return False

        if '/' in template_name:
            self.send(src_addr, "/error", path, ray.Err.CREATE_FAILED,
                      "Invalid session name.")
            return False

    @ray_method('/ray/server/get_session_preview', 's')
    def rayServerGetSessionPreview(self, path, args, types, src_addr):
        self.session_to_preview = args[0]
    
    @ray_method('/ray/server/script_info', 's')
    def rayServerScriptInfo(self, path, args, types, src_addr):
        self.send_gui('/ray/gui/script_info', args[0])
        self.send(src_addr, "/reply", path, "Info sent")

    @ray_method('/ray/server/hide_script_info', '')
    def rayServerHideScriptInfo(self, path, args, types, src_addr):
        self.send_gui('/ray/gui/hide_script_info')
        self.send(src_addr, "/reply", path, "Info hidden")

    @ray_method('/ray/server/script_user_action', 's')
    def rayServerScriptUserAction(self, path, args, types, src_addr):
        if not self.gui_list:
            self.send(src_addr, '/error', path, ray.Err.LAUNCH_FAILED,
                      "This server has no attached GUI")
            return
        self.send_gui('/ray/gui/script_user_action', args[0])

    # set option from GUI
    @ray_method('/ray/server/set_option', 'i')
    def rayServerSetOption(self, path, args, types, src_addr):
        option = args[0]
        self._set_option(option)

        for gui_addr in self.gui_list:
            if not ray.are_same_osc_port(gui_addr.url, src_addr.url):
                self.send(gui_addr, '/ray/gui/server/options', self.options)

    # set options from ray_control
    @ray_method('/ray/server/set_options', None)
    def rayServerSetOptions(self, path, args, types, src_addr: liblo.Address):
        if not ray.types_are_all_strings(types):
            self._unknown_message(path, types, src_addr)
            return False

        if TYPE_CHECKING:
            assert isinstance(args, list[str])

        for option_str in args:
            option_value = True
            if option_str.startswith('not_'):
                option_value = False
                option_str = option_str.replace('not_', '', 1)

            if option_str in self.options_dict:
                option = self.options_dict[option_str]
                if option_value:
                    if (option == ray.Option.DESKTOPS_MEMORY
                            and not self.options & ray.Option.HAS_WMCTRL):
                        self.send(src_addr, '/minor_error', path,
                            "wmctrl is not present. "
                            "Impossible to activate 'desktops_memory' option")
                        continue
                    if (option == ray.Option.SNAPSHOTS
                            and not self.options & ray.Option.HAS_GIT):
                        self.send(src_addr, '/minor_error', path,
                            "git is not present. "
                            "Impossible to activate 'snapshots' option")
                        continue

                if not option_value:
                    option = -option
                self._set_option(option)

        for gui_addr in self.gui_list:
            if not ray.are_same_osc_port(gui_addr.url, src_addr.url):
                self.send(gui_addr, '/ray/gui/server/options', self.options)

        self.send(src_addr, '/reply', path, 'Options set')

    @ray_method('/ray/server/has_option', 's')
    def rayServerHasOption(self, path, args, types, src_addr):
        option_str = args[0]
        option_value = False

        if option_str not in self.options_dict:
            self.send(src_addr, '/error', path, ray.Err.GENERAL_ERROR,
                      "option \"%s\" doesn't exists" % option_str)
            return

        if self.options & self.options_dict[option_str]:
            self.send(src_addr, '/reply', path, 'Has option')
        else:
            self.send(src_addr, '/error', path, ray.Err.GENERAL_ERROR,
                      "Option %s is not currently used" % option_str)

    @ray_method('/ray/server/clear_client_templates_database', '')
    def rayServerClearClientTemplatesDatabase(self, path, args, types, src_addr):
        self.client_templates_database['factory'].clear()
        self.client_templates_database['user'].clear()
        self.send(src_addr, '/reply', path, 'database cleared')

    @ray_method('/ray/server/open_file_manager_at', 's')
    def rayServerOpenFileManagerAt(self, path, args, types, src_addr):
        folder_path = args[0]
        if os.path.isdir(folder_path):
            subprocess.Popen(['xdg-open', folder_path])
        self.send(src_addr, '/reply', path, '')

    @ray_method('/ray/server/exotic_action', 's')
    def rayServerExoticAction(self, path, args, types, src_addr):
        action = args[0]
        autostart_dir = Path.home() / '.config' / 'autostart'
        desk_file = "ray-jack_checker.desktop"
        dest_full_path = autostart_dir / desk_file

        if action == 'set_jack_checker_autostart':
            if not autostart_dir.exists():
                autostart_dir.mkdir(parents=True)

            src_full_file = (
                get_code_root()
                / 'data' / 'share' / 'applications' / desk_file)
            
            shutil.copyfile(src_full_file, dest_full_path)

        elif action == 'unset_jack_checker_autostart':
            dest_full_path.unlink(missing_ok=True)

    @ray_method('/ray/server/patchbay/save_group_position',
                ray.GroupPosition.sisi())
    def rayServerPatchbaySaveCoordinates(self, path, args, types, src_addr):
        # here send to others GUI the new group position
        for gui_addr in self.gui_list:
            if not ray.are_same_osc_port(gui_addr.url, src_addr.url):
                self.send(gui_addr, '/ray/gui/patchbay/update_group_position',
                          *args)

    @ray_method('/ray/server/patchbay/save_portgroup', None)
    def rayServerPatchbaySavePortGroup(self, path, args, types: str, src_addr):
        # args must be group_name, port_type, port_mode, above_metadatas, *port_names
        # where port_names are all strings
        # so types must start with 'siiis' and may continue with strings only
        if not types.startswith('siiis'):
            self._unknown_message(path, types, src_addr)
            return False

        other_types = types.replace('siiis', '', 1)
        for t in other_types:
            if t != 's':
                self._unknown_message(path, types, src_addr)
                return False

    @ray_method('/ray/session/save', '')
    def raySessionSave(self, path, args, types, src_addr):
        if not self.session.path:
            self.send(src_addr, "/error", path, ray.Err.NO_SESSION_OPEN,
                      "No session to save.")
            return False

    @ray_method('/ray/session/run_step', None)
    def raySessionProcessStep(self, path, args, types, src_addr):
        if not ray.types_are_all_strings(types):
            self._unknown_message(path, types, src_addr)
            return False

    @ray_method('/ray/session/save_as_template', 's')
    def raySessionSaveAsTemplate(self, path, args, types, src_addr):
        template_name = args[0]
        if '/' in template_name or template_name == '.':
            self.send(src_addr, "/error", path, ray.Err.CREATE_FAILED,
                      "Invalid session template name.")
            return False

    @ray_method('/ray/session/get_session_name', '')
    def raySessionGetSessionName(self, path, args, types, src_addr):
        if not self.session.path:
            self.send(src_addr, "/error", path, ray.Err.NO_SESSION_OPEN,
                      "No session loaded.")
            return False

        self.send(src_addr, '/reply', path, self.session.name)
        self.send(src_addr, '/reply', path)
        return False

    @ray_method('/ray/session/take_snapshot', 's')
    def raySessionTakeSnapshotOnly(self, path, args, types, src_addr):
        if not self.options & ray.Option.HAS_GIT:
            self.send(src_addr, '/error', path,
                      "snapshot impossible because git is not installed")
            return False

    @ray_method('/ray/session/take_snapshot', 'si')
    def raySessionTakeSnapshot(self, path, args, types, src_addr):
        if not self.options & ray.Option.HAS_GIT:
            self.send(src_addr, '/error', path,
                      "snapshot impossible because git is not installed")
            return False

    @ray_method('/ray/session/close', '')
    def raySessionClose(self, path, args, types, src_addr):
        if not self.session.path:
            self.send(src_addr, "/error", path, ray.Err.NO_SESSION_OPEN,
                      "No session to close.")
            return False

    @ray_method('/ray/session/abort', '')
    def raySessionAbort(self, path, args, types, src_addr):
        pass

    @ray_method('/ray/session/cancel_close', '')
    def raySessionCancelClose(self, path, args, types, src_addr):
        if not self.session.path:
            self.send(src_addr, "/error", path, ray.Err.NO_SESSION_OPEN,
                      "No session to cancel close.")
            return False

    @ray_method('/ray/session/skip_wait_user', '')
    def raySessionSkipWaitUser(self, path, args, types, src_addr):
        if self.server_status != ray.ServerStatus.WAIT_USER:
            return False

    @ray_method('/ray/session/duplicate', 's')
    def raySessionDuplicate(self, path, args, types, src_addr):
        if self.is_nsm_locked:
            return False

        if not self.session.path:
            self.send(src_addr, "/error", path, ray.Err.NO_SESSION_OPEN,
                      "No session to duplicate.")
            return False

        if not _path_is_valid(args[0]):
            self.send(src_addr, "/error", path, ray.Err.CREATE_FAILED,
                      "Invalid session name.")
            return False

    @ray_method('/ray/session/duplicate_only', 'sss')
    def nsmServerDuplicateOnly(self, path, args, types, src_addr):
        self.send(src_addr, '/ray/net_daemon/duplicate_state', 0)

    @ray_method('/ray/session/open_snapshot', 's')
    def raySessionOpenSnapshot(self, path, args, types, src_addr):
        pass

    @ray_method('/ray/session/rename', 's')
    def rayServerRename(self, path, args, types, src_addr):
        new_session_name = args[0]

        #prevent rename session in network session
        if self._nsm_locker_url:
            NSM_URL = os.getenv('NSM_URL')
            if not NSM_URL:
                return False

            if not ray.are_same_osc_port(self._nsm_locker_url, NSM_URL):
                return False

        if '/' in new_session_name:
            self.send(src_addr, "/error", path, ray.Err.CREATE_FAILED,
                      "Invalid session name.")
            return False

        if self._is_operation_pending(src_addr, path):
            return False

        if not self.session.path:
            self.send(src_addr, "/error", path, ray.Err.NO_SESSION_OPEN,
                      "No session to rename.")
            return False

    @ray_method('/ray/session/set_notes', 's')
    def raySessionSetNotes(self, path, args, types, src_addr):
        self.session.notes = args[0]

        for gui_addr in self.gui_list:
            if not ray.are_same_osc_port(gui_addr.url, src_addr.url):
                self.send(gui_addr, '/ray/gui/session/notes',
                          self.session.notes)

    @ray_method('/ray/session/get_notes', '')
    def raySessionGetNotes(self, path, args, types, src_addr):
        pass

    @ray_method('/ray/session/add_executable', 'siiissi')
    def raySessionAddExecutableAdvanced(self, path, args, types, src_addr):
        executable_path, auto_start, protocol, \
            prefix_mode, prefix_pattern, client_id, jack_naming = args

        if not self.session.path:
            self.send(src_addr, "/error", path, ray.Err.NO_SESSION_OPEN,
                      "Cannot add to session because no session is loaded.")
            return False

        if protocol == ray.Protocol.NSM and '/' in executable_path:
            self.send(src_addr, "/error", path, ray.Err.LAUNCH_FAILED,
                "Absolute paths are not permitted. Clients must be in $PATH")
            return False

    @ray_method('/ray/session/add_executable', None)
    def raySessionAddExecutableStrings(self, path, args, types, src_addr):
        if not (types and ray.types_are_all_strings(types)):
            self._unknown_message(path, types, src_addr)
            return False

        if not self.session.path:
            self.send(src_addr, "/error", path, ray.Err.NO_SESSION_OPEN,
                      "Cannot add to session because no session is loaded.")
            return False

        executable_path = args[0]
        via_proxy = bool(len(args) > 1 and 'via_proxy' in args[1:])
        ray_hack = bool(len(args) > 1 and 'ray_hack' in args[1:])

        if '/' in executable_path and not (via_proxy or ray_hack):
            self.send(src_addr, "/error", path, ray.Err.LAUNCH_FAILED,
                "Absolute paths are not permitted. Clients must be in $PATH")
            return False

    @ray_method('/ray/session/add_client_template', 'isss')
    def rayServerAddClientTemplate(self, path, args, types, src_addr):
        pass

    @ray_method('/ray/session/add_factory_client_template', None)
    def raySessionAddFactoryClientTemplate(self, path, args, types, src_addr):
        if not (types and ray.types_are_all_strings(types)):
            self._unknown_message(path, types, src_addr)
            return False

    @ray_method('/ray/session/add_user_client_template', None)
    def raySessionAddUserClientTemplate(self, path, args, types, src_addr):
        if not (types and ray.types_are_all_strings(types)):
            self._unknown_message(path, types, src_addr)
            return False

    @ray_method('/ray/session/add_other_session_client', 'ss')
    def raySessionEatOtherSessionClient(self, path, args, types, src_addr):
        pass

    @ray_method('/ray/session/reorder_clients', None)
    def rayServerReorderClients(self, path, args, types, src_addr):
        if not (types and ray.types_are_all_strings(types)):
            self._unknown_message(path, types, src_addr)
            return False

    @ray_method('/ray/session/list_snapshots', '')
    def rayServerListSnapshots(self, path, args, types, src_addr):
        pass

    @ray_method('/ray/session/set_auto_snapshot', 'i')
    def rayServerSetAutoSnapshot(self, path, args, types, src_addr):
        pass

    @ray_method('/ray/session/open_folder', '')
    def rayServerOpenFolder(self, path, args, types, src_addr):
        if self.session.path:
            subprocess.Popen(['xdg-open', self.session.path])
        self.send(src_addr, '/reply', path, '')

    @ray_method('/ray/session/clear_clients', None)
    def raySessionStopClients(self, path, args, types, src_addr):
        if not ray.types_are_all_strings(types):
            self._unknown_message(path, types, src_addr)
            return False

    @ray_method('/ray/session/show_notes', '')
    def raySessionShowNotes(self, path, args, types, src_addr):
        if not self.session.path:
            self.send(src_addr, '/error', path, ray.Err.NO_SESSION_OPEN,
                      "No session to show notes")
            return False

        self.session.notes_shown = True
        self.send_gui('/ray/gui/session/notes_shown')
        self.send(src_addr, '/reply', path, 'notes shown')

    @ray_method('/ray/session/hide_notes', '')
    def raySessionHideNotes(self, path, args, types, src_addr):
        if not self.session.path:
            self.send(src_addr, '/error', path, ray.Err.NO_SESSION_OPEN,
                      "No session to hide notes")
            return False

        self.session.notes_shown = False
        self.send_gui('/ray/gui/session/notes_hidden')
        self.send(src_addr, '/reply', path, 'notes hidden')

    @ray_method('/ray/session/list_clients', None)
    def raySessionListClients(self, path, args, types, src_addr):
        if not ray.types_are_all_strings(types):
            self._unknown_message(path, types, src_addr)
            return False

    @ray_method('/ray/session/list_trashed_clients', '')
    def raySessionListTrashedClients(self, path, args, types, src_addr):
        pass

    @ray_method('/ray/client/stop', 's')
    def rayGuiClientStop(self, path, args, types, src_addr):
        pass

    @ray_method('/ray/client/kill', 's')
    def rayGuiClientKill(self, path, args, types, src_addr):
        pass

    @ray_method('/ray/client/trash', 's')
    def rayGuiClientRemove(self, path, args, types, src_addr):
        pass

    @ray_method('/ray/client/start', 's')
    def rayGuiClientStart(self, path, args, types, src_addr):
        pass

    @ray_method('/ray/client/resume', 's')
    def rayGuiClientResume(self, path, args, types, src_addr):
        pass

    @ray_method('/ray/client/open', 's')
    def rayClientOpen(self, path, args, types, src_addr):
        pass

    @ray_method('/ray/client/save', 's')
    def rayGuiClientSave(self, path, args, types, src_addr):
        pass

    @ray_method('/ray/client/save_as_template', 'ss')
    def rayGuiClientSaveAsTemplate(self, path, args, types, src_addr):
        pass

    @ray_method('/ray/client/show_optional_gui', 's')
    def nsmGuiClientShow_optional_gui(self, path, args, types, src_addr):
        pass

    @ray_method('/ray/client/hide_optional_gui', 's')
    def nsmGuiClientHide_optional_gui(self, path, args, types, src_addr):
        pass

    @ray_method('/ray/client/update_properties', ray.ClientData.sisi())
    def rayGuiClientUpdateProperties(self, path, args, types, src_addr):
        pass

    @ray_method('/ray/client/update_ray_hack_properties', 's' + ray.RayHack.sisi())
    def rayClientUpdateRayHackProperties(self, path, args, types, src_addr):
        pass

    @ray_method('/ray/client/update_ray_net_properties', 's' + ray.RayNet.sisi())
    def rayClientUpdateRayNetProperties(self, path, args, types, src_addr):
        pass

    @ray_method('/ray/client/get_properties', 's')
    def rayClientGetProperties(self, path, args, types, src_addr):
        pass

    @ray_method('/ray/client/set_properties', None)
    def rayGuiClientSetProperties(self, path, args, types, src_addr):
        if not (len(args) >= 2 and ray.types_are_all_strings(types)):
            self._unknown_message(path, types, src_addr)
            return

    @ray_method('/ray/client/get_proxy_properties', 's')
    def rayClientGetProxyProperties(self, path, args, types, src_addr):
        pass

    @ray_method('/ray/client/set_proxy_properties', None)
    def rayClientSetProxyProperties(self, path, args, types, src_addr):
        if not (len(args) >= 2 and ray.types_are_all_strings(types)):
            self._unknown_message(path, types, src_addr)
            return

    @ray_method('/ray/client/change_prefix', None)
    def rayClientChangePrefix(self, path, args, types, src_addr):
        # here message can be si, ss, sis, sss
        invalid = False

        if len(args) < 2:
            invalid = True

        elif args[1] in (ray.PrefixMode.CUSTOM, 'custom'):
            if len(args) < 3:
                invalid = True

        if invalid:
            self._unknown_message(path, types, src_addr)
            return False

    @ray_method('/ray/client/change_advanced_properties', 'ssisi')
    def rayClientChangeAdvancedProperties(self, path, args, types, src_addr):
        pass
    
    @ray_method('/ray/client/full_rename', 'ss')
    def rayClientFullRename(self, path, args, types, src_addr):
        pass
    
    @ray_method('/ray/client/change_id', 'ss')
    def rayClientChangeId(self, path, args, types, src_addr):
        pass

    @ray_method('/ray/client/set_description', 'ss')
    def rayClientSetDescription(self, path, args, types, src_addr):
        pass

    @ray_method('/ray/client/get_description', 's')
    def rayClientGetDescription(self, path, args, types, src_addr):
        pass

    @ray_method('/ray/client/get_pid', 's')
    def ratClientGetPid(self, path, args, types, src_addr):
        pass

    @ray_method('/ray/client/list_files', 's')
    def rayClientListFiles(self, path, args, types, src_addr):
        pass

    @ray_method('/ray/client/list_snapshots', 's')
    def rayClientListSnapshots(self, path, args, types, src_addr):
        pass

    @ray_method('/ray/client/open_snapshot', 'ss')
    def rayClientLoadSnapshot(self, path, args, types, src_addr):
        pass

    @ray_method('/ray/client/is_started', 's')
    def rayClientIsStarted(self, path, args, types, src_addr):
        pass

    @ray_method('/ray/client/set_custom_data', 'sss')
    def rayClientSetCustomData(self, path, args, types, src_addr):
        pass

    @ray_method('/ray/client/get_custom_data', 'ss')
    def rayClientGetCustomData(self, path, args, types, src_addr):
        pass

    @ray_method('/ray/client/set_tmp_data', 'sss')
    def rayClientSetTmpData(self, path, args, types, src_addr):
        pass

    @ray_method('/ray/client/get_tmp_data', 'ss')
    def rayClientGetTmpData(self, path, args, types, src_addr):
        pass

    @ray_method('/ray/client/send_signal', 'si')
    def rayClientSendSignal(self, path, args, types, src_addr):
        pass

    @ray_method('/ray/trashed_client/restore', 's')
    def rayTrashedClientRestore(self, path, args, types, src_addr):
        pass

    @ray_method('/ray/trashed_client/remove_definitely', 's')
    def rayTrashedClientRemoveDefinitely(self, path, args, types, src_addr):
        pass

    @ray_method('/ray/trashed_client/remove_keep_files', 's')
    def rayTrashedClientRemoveKeepFiles(self, path, args, types, src_addr):
        pass

    @ray_method('/ray/net_daemon/duplicate_state', 'f')
    def rayDuplicateState(self, path, args, types, src_addr):
        pass

    @ray_method('/ray/favorites/add', 'ssis')
    def rayFavoriteAdd(self, path, args, types, src_addr):
        name, icon, int_factory, display_name = args

        for favorite in RS.favorites:
            if (favorite.name == name
                    and bool(int_factory) == favorite.factory):
                favorite.icon = icon
                favorite.display_name = display_name
                break
        else:
            RS.favorites.append(
                ray.Favorite(name, icon, bool(int_factory), display_name))

        self.send_gui('/ray/gui/favorites/added', *args)

    @ray_method('/ray/favorites/remove', 'si')
    def rayFavoriteRemove(self, path, args, types, src_addr):
        name, int_factory = args

        for favorite in RS.favorites:
            if (favorite.name == name
                    and bool(int_factory) == favorite.factory):
                RS.favorites.remove(favorite)
                break

        self.send_gui('/ray/gui/favorites/removed', *args)

    @ray_method(None, None)
    def noneMethod(self, path, args, types, src_addr):
        types_str = ''
        for t in types:
            types_str += t

        self._unknown_message(path, types, src_addr)
        return False

    def _is_operation_pending(self, src_addr, path):
        if self.session.file_copier.is_active():
            self.send(src_addr, "/error", path, ray.Err.COPY_RUNNING,
                      "ray-daemon is copying files. "
                        + "Wait copy finish or abort copy, "
                        + "and restart operation !")
            return True

        if self.session.steps_order:
            self.send(src_addr, "/error", path, ray.Err.OPERATION_PENDING,
                      "An operation pending.")
            return True

        return False

    def _set_option(self, option: int):
        if option >= 0:
            self.options |= option
        else:
            self.options &= ~abs(option)

    def send(self, *args):
        if CommandLineArgs.debug:
            sys.stderr.write(
                '\033[96mOSC::daemon sends\033[0m ' + str(args[1:]) + '\n')

        ClientCommunicating.send(self, *args)

    def send_gui(self, *args):
        for gui_addr in self.gui_list:
            self.send(gui_addr, *args)

    def set_server_status(self, server_status:int):
        self.server_status = server_status
        self.send_gui('/ray/gui/server/status', server_status)

    def send_renameable(self, renameable:bool):
        if not renameable:
            self.send_gui('/ray/gui/session/renameable', 0)
            return

        if self._nsm_locker_url:
            nsm_url = os.getenv('NSM_URL')
            if not nsm_url:
                return
            if not ray.are_same_osc_port(self._nsm_locker_url, nsm_url):
                return

        self.send_gui('/ray/gui/session/renameable', 1)

    def announce_gui(self, url, nsm_locked=False, is_net_free=True, gui_pid=0):
        gui_addr = GuiAdress(url)
        gui_addr.gui_pid = gui_pid

        self.send(gui_addr, "/ray/gui/server/announce", ray.VERSION,
                  self.server_status, self.options, self.session.root,
                  int(is_net_free))

        self.send(gui_addr, "/ray/gui/server/status", self.server_status)
        self.send(gui_addr, "/ray/gui/session/name",
                  self.session.name, self.session.path)
        self.send(gui_addr, '/ray/gui/session/notes', self.session.notes)
        self.send(gui_addr, '/ray/gui/server/terminal_command',
                  self.terminal_command)

        self.session.canvas_saver.send_all_group_positions(gui_addr)

        for favorite in RS.favorites:
            self.send(gui_addr, "/ray/gui/favorites/added",
                      favorite.name, favorite.icon, int(favorite.factory),
                      favorite.display_name)

        for client in self.session.clients:
            self.send(gui_addr,
                      '/ray/gui/client/new',
                      *client.spread())

            if client.protocol == ray.Protocol.RAY_HACK:
                self.send(gui_addr,
                          '/ray/gui/client/ray_hack_update',
                          client.client_id,
                          *client.ray_hack.spread())
            elif client.protocol == ray.Protocol.RAY_NET:
                self.send(gui_addr,
                          '/ray/gui/client/ray_net_update',
                          client.client_id,
                          *client.ray_net.spread())

            self.send(gui_addr, "/ray/gui/client/status",
                      client.client_id, client.status)

            if client.is_capable_of(':optional-gui:'):
                self.send(gui_addr, '/ray/gui/client/gui_visible',
                          client.client_id, int(client.gui_visible))

            if client.is_capable_of(':dirty:'):
                self.send(gui_addr, '/ray/gui/client/dirty',
                          client.client_id, client.dirty)

        for trashed_client in self.session.trashed_clients:
            self.send(gui_addr, '/ray/gui/trash/add',
                      *trashed_client.spread())

            if trashed_client.protocol == ray.Protocol.RAY_HACK:
                self.send(gui_addr, '/ray/gui/trash/ray_hack_update',
                          trashed_client.client_id,
                          *trashed_client.ray_hack.spread())
            elif trashed_client.protocol == ray.Protocol.RAY_NET:
                self.send(gui_addr, '/ray/gui/trash/ray_net_update',
                          trashed_client.client_id,
                          *trashed_client.ray_net.spread())

        self.session.check_recent_sessions_existing()
        if self.session.root in self.session.recent_sessions.keys():
            self.send(gui_addr, '/ray/gui/server/recent_sessions',
                      *self.session.recent_sessions[self.session.root])

        self.send(gui_addr, '/ray/gui/server/message',
                  _translate('daemon', "daemon runs at %s") % self.url)

        self.gui_list.append(gui_addr)

        multi_daemon_file = MultiDaemonFile.get_instance()
        if multi_daemon_file:
            multi_daemon_file.update()

        Terminal.message("GUI connected at %s" % gui_addr.url)

    def announce_controller(self, control_address):
        controller = Controller()
        controller.addr = control_address
        self.controller_list.append(controller)
        self.send(control_address, "/ray/control/server/announce",
                  ray.VERSION, self.server_status, self.options,
                  self.session.root, 1)

    def send_controller_message(self, message):
        for controller in self.controller_list:
            self.send(controller.addr, '/ray/control/message', message)

    def has_gui(self)->int:
        has_gui = False

        for gui_addr in self.gui_list:
            if ray.are_on_same_machine(self.url, gui_addr.url):
                # we've got a local GUI
                return 3

            has_gui = True

        if has_gui:
            return 1

        return 0

    def get_local_gui_pid_list(self) -> str:
        pid_list = []
        for gui_addr in self.gui_list:
            if ray.are_on_same_machine(gui_addr.url, self.url):
                pid_list.append(str(gui_addr.gui_pid))
        return ':'.join(pid_list)

    def is_gui_address(self, addr: liblo.Address) -> bool:
        for gui_addr in self.gui_list:
            if ray.are_same_osc_port(gui_addr.url, addr.url):
                return True
        return False




import time
import tempfile
import socket
import json
import subprocess
from typing import TYPE_CHECKING

try:
    from liblo import Server, Address
except ImportError:
    from pyliblo3 import Server, Address

if TYPE_CHECKING:
    from ray_jackpatch_to_osc import MainObject, TransportPosition

### Code copied from shared/ray.py
### we don't import ray.py here, because this executable is Qt free
### TODO : make a miniray.py with only Qt free code

class Machine192:
    ip = ''
    read_done = False
    
    @staticmethod
    def read()->str:
        try:
            ips = subprocess.check_output(
                ['ip', 'route', 'get', '1']).decode()
            ip_line = ips.partition('\n')[0]
            ip_end = ip_line.rpartition('src ')[2]
            ip = ip_end.partition(' ')[0]

        except BaseException:
            try:
                ips = subprocess.check_output(['hostname', '-I']).decode()
                ip = ips.split(' ')[0]
            except BaseException:
                return ''

        if ip.count('.') != 3:
            return ''
        
        return ip
    
    @classmethod
    def get(cls)->str:
        if cls.read_done:
            return cls.ip
        
        cls.ip = cls.read()
        cls.read_done = True

        return cls.ip

def areOnSameMachine(url1: str, url2: str):
    if url1 == url2:
        return True

    try:
        address1 = Address(url1)
        address2 = Address(url2)
    except BaseException:
        return False

    if address1.hostname == address2.hostname:
        return True

    try:
        if (socket.gethostbyname(address1.hostname)
                    in ('127.0.0.1', '127.0.1.1')
                and socket.gethostbyname(address2.hostname)
                    in ('127.0.0.1', '127.0.1.1')):
            return True

        if socket.gethostbyaddr(
                address1.hostname) == socket.gethostbyaddr(
                address2.hostname):
            return True

    except BaseException:
        try:
            ip = Machine192.get()

            if ip not in (address1.hostname, address2.hostname):
                return False

            try:
                if socket.gethostbyname(
                        address1.hostname) in (
                        '127.0.0.1',
                        '127.0.1.1'):
                    if address2.hostname == ip:
                        return True
            except BaseException:
                if socket.gethostbyname(
                        address2.hostname) in (
                        '127.0.0.1',
                        '127.0.1.1'):
                    if address1.hostname == ip:
                        return True

        except BaseException:
            return False

        return False

    return False


class OscJackPatch(Server):
    SLOW_WAIT_TIME = 0.020
    SLOW_WAIT_NUM = 50
    
    def __init__(self, main_object: 'MainObject'):
        Server.__init__(self)
        self.add_method('/ray/patchbay/add_gui', 's',
                        self._ray_patchbay_add_gui)
        self.add_method('/ray/patchbay/gui_disannounce', '',
                        self._ray_patchbay_gui_disannounce)
        self.add_method('ray/patchbay/port/set_alias', 'sis',
                        self._ray_patchbay_port_set_alias)
        self.add_method('/ray/patchbay/connect', 'ss',
                        self._ray_patchbay_connect)
        self.add_method('/ray/patchbay/disconnect', 'ss',
                        self._ray_patchbay_disconnect)
        self.add_method('/ray/patchbay/set_buffer_size', 'i',
                        self._ray_patchbay_set_buffersize)
        self.add_method('/ray/patchbay/refresh', '',
                        self._ray_patchbay_refresh)
        self.add_method('/ray/patchbay/set_metadata', 'hss',
                        self._ray_patchbay_set_metadata)
        self.add_method('/ray/patchbay/transport_play', 'i',
                        self._ray_patchbay_transport_play)
        self.add_method('/ray/patchbay/transport_stop', '',
                         self._ray_patchbay_transport_stop)
        self.add_method('/ray/patchbay/transport_relocate', 'i',
                        self._ray_patchbay_transport_relocate)
        self.add_method('/ray/patchbay/activate_dsp_load', 'i',
                        self._ray_patchbay_activate_dsp_load)
        self.add_method('/ray/patchbay/activate_transport', 'i',
                        self._ray_patchbay_activate_transport)

        self.main_object = main_object
        self.port_list = main_object.port_list
        self.connection_list = main_object.connection_list
        self.metadata_list = main_object.metadata_list
        self.client_list = main_object.client_list
        self.gui_list = list[Address]()
        self._tmp_gui_url = ''
        self._terminate = False

    def set_tmp_gui_url(self, gui_url: str):
        self._tmp_gui_url = gui_url
    
    def _ray_patchbay_add_gui(self, path, args, types, src_addr):
        self.add_gui(args[0])

    def _ray_patchbay_gui_disannounce(self, path, args, types, src_addr):
        for gui_addr in self.gui_list:
            if gui_addr.url == src_addr.url:
                # possible because we break the loop
                self.gui_list.remove(gui_addr)
                break
        
        if not self.gui_list:
            # no more GUI connected, no reason to exists anymore
            self._terminate = True

    def _ray_patchbay_port_set_alias(self, path, args, types, src_addr):
        port_name, alias_num, alias = args
        for port in self.port_list:
            if port.name == port_name:
                # TODO
                # better would be to use jacklib.port_set_alias(port, alias)
                # but this is very confuse
                # 2 aliases possibles, but only one arg to this method (after port).
                if alias_num == 1:
                    port.alias_1 = alias
                elif alias_num == 2:
                    port.alias_2 = alias
                break

    def _ray_patchbay_connect(self, path, args):
        port_out_name, port_in_name = args
        # connect here
        self.main_object.connect_ports(port_out_name, port_in_name)
    
    def _ray_patchbay_disconnect(self, path, args):
        port_out_name, port_in_name = args
        # disconnect here
        self.main_object.connect_ports(
            port_out_name, port_in_name, disconnect=True)

    def _ray_patchbay_set_buffersize(self, path, args):
        buffer_size = args[0]
        self.main_object.set_buffer_size(buffer_size)

    def _ray_patchbay_refresh(self, path, args):
        self.main_object.refresh()

    def _ray_patchbay_set_metadata(self, path, args):
        uuid, key, value = args
        self.main_object.set_metadata(uuid, key, value)

    def _ray_patchbay_transport_play(self, path, args):
        self.main_object.transport_play(bool(args[0]))
    
    def _ray_patchbay_transport_stop(self, path, args):
        self.main_object.transport_stop()
    
    def _ray_patchbay_transport_relocate(self, path, args):
        self.main_object.transport_relocate(args[0])

    def _ray_patchbay_activate_dsp_load(self, path, args):
        self.main_object.dsp_wanted = bool(args[0])
        
    def _ray_patchbay_activate_transport(self, path, args):
        self.main_object.set_transport_wanted(args[0])

    def send_gui(self, *args):
        for gui_addr in self.gui_list:
            self.send(gui_addr, *args)

    def multi_send(self, src_addr_list, *args):
        for src_addr in src_addr_list:
            self.send(src_addr, *args)

    def send_local_data(self, src_addr_list):
        # at invitation, if gui is on the same machine
        # it's prefferable to save all data in /tmp
        # Indeed, to prevent OSC packet loses
        # this daemon will send a lot of OSC messages not too fast
        # so here, it is faster, and prevent OSC saturation.
        # json format (and not binary with pickle) is choosen
        # this way, code language of the GUI is not a blocker
        patchbay_data = {'ports': list[dict](), 'connections': list[dict](),
                         'metadatas': list[dict](), 'clients': list[dict]()}
        for port in self.port_list:
            port_dict = {'name': port.name, 'type': port.type,
                         'flags': port.flags, 'uuid': port.uuid}
            patchbay_data['ports'].append(port_dict)
        
        for connection in self.connection_list:
            conn_dict = {'port_out_name': connection[0],
                         'port_in_name': connection[1]}
            patchbay_data['connections'].append(conn_dict)

        for metadata in self.metadata_list:
            patchbay_data['metadatas'].append(metadata)

        for client_dict in self.client_list:
            patchbay_data['clients'].append(client_dict)

        if self.main_object.alsa_mng is not None:
            alsa_mng = self.main_object.alsa_mng
            
            for pdata in alsa_mng.parse_ports_and_flags():
                port_dict = {'name': pdata.name, 'type': pdata.type,
                             'flags': pdata.flags, 'uuid': pdata.uuid}
                patchbay_data['ports'].append(port_dict)
                
            for port_out_name, port_in_name in alsa_mng.parse_connections():
                conn_dict = {'port_out_name': port_out_name,
                             'port_in_name': port_in_name}
                patchbay_data['connections'].append(conn_dict)

        for src_addr in src_addr_list:
            # tmp file is deleted by the gui itself once read
            # so there is one tmp file per local GUI
            file = tempfile.NamedTemporaryFile(delete=False, mode='w+')
            json.dump(patchbay_data, file)
            file.close()

            self.send(src_addr, '/ray/gui/patchbay/fast_temp_file_running',
                      file.name)

    def send_distant_data(self, src_addr_list):
        # we need to slow the long process of messages sends
        # to prevent loss packets
        self.multi_send(src_addr_list, '/ray/gui/patchbay/big_packets', 0)
        n = 0
        increment = len(src_addr_list)

        for port in self.port_list:
            self.multi_send(src_addr_list, '/ray/gui/patchbay/port_added',
                            port.name, port.type, port.flags, port.uuid)
            
            n += increment
            if n % self.SLOW_WAIT_NUM < increment:
                self.multi_send(src_addr_list,
                                '/ray/gui/patchbay/big_packets', 1)
                time.sleep(self.SLOW_WAIT_TIME)
                self.multi_send(src_addr_list,
                                '/ray/gui/patchbay/big_packets', 0)            

        for connection in self.connection_list:
            self.multi_send(src_addr_list,
                            '/ray/gui/patchbay/connection_added',
                            connection[0], connection[1])
            
            n += increment
            if n % self.SLOW_WAIT_NUM < increment:
                self.multi_send(src_addr_list,
                                '/ray/gui/patchbay/big_packets', 1)
                time.sleep(self.SLOW_WAIT_TIME)
                self.multi_send(src_addr_list,
                                '/ray/gui/patchbay/big_packets', 0)
                
        for metadata in self.metadata_list:
            self.multi_send(src_addr_list,
                            '/ray/gui/patchbay/metadata_updated',
                            metadata['uuid'], metadata['key'],
                            metadata['value'])
            
            n += increment
            if n % self.SLOW_WAIT_NUM < increment:
                self.multi_send(src_addr_list,
                                '/ray/gui/patchbay/big_packets', 1)
                time.sleep(self.SLOW_WAIT_TIME)
                self.multi_send(src_addr_list,
                                '/ray/gui/patchbay/big_packets', 0)

        if self.main_object.alsa_mng is not None:
            alsa_mng = self.main_object.alsa_mng
            for port in alsa_mng.parse_ports_and_flags():
                self.multi_send(src_addr_list, '/ray/gui/patchbay/port_added',
                                port.name, port.type, port.flags, port.uuid)

                n += increment
                if n % self.SLOW_WAIT_NUM < increment:
                    self.multi_send(src_addr_list,
                                    '/ray/gui/patchbay/big_packets', 1)
                    time.sleep(self.SLOW_WAIT_TIME)
                    self.multi_send(src_addr_list,
                                    '/ray/gui/patchbay/big_packets', 0)
            
            for conn in alsa_mng.parse_connections():
                self.multi_send(
                    src_addr_list,
                    '/ray/gui/patchbay/connection_added',
                    *conn)

                n += increment
                if n % self.SLOW_WAIT_NUM < increment:
                    self.multi_send(src_addr_list,
                                    '/ray/gui/patchbay/big_packets', 1)
                    time.sleep(self.SLOW_WAIT_TIME)
                    self.multi_send(src_addr_list,
                                    '/ray/gui/patchbay/big_packets', 0)

        self.multi_send(src_addr_list, '/ray/gui/patchbay/big_packets', 1)

    def add_gui(self, gui_url):
        gui_addr = Address(gui_url)
        if gui_addr is None:
            return
        
        self.send(gui_addr, '/ray/gui/patchbay/announce',
                  int(self.main_object.jack_running),
                  self.main_object.samplerate,
                  self.main_object.buffer_size)

        self.send(gui_addr, '/ray/gui/patchbay/dsp_load',
                  self.main_object.last_sent_dsp_load)

        tpos = self.main_object.last_transport_pos
        self.send(gui_addr, '/ray/gui/patchbay/transport_position',
                  tpos.frame, int(tpos.rolling), int(tpos.valid_bbt),
                  tpos.bar, tpos.beat, tpos.tick, tpos.beats_per_minutes)

        if areOnSameMachine(gui_url, self.url):
            self.send_local_data([gui_addr])
        else:
            self.send_distant_data([gui_addr])
        
        self.gui_list.append(gui_addr)

    def server_restarted(self):
        self.send_gui('/ray/gui/patchbay/server_started')
        self.send_samplerate()
        self.send_buffersize()
        
        local_guis = []
        distant_guis = []
        
        for gui_addr in self.gui_list:
            if areOnSameMachine(self.url, gui_addr.url):
                local_guis.append(gui_addr)
            else:
                distant_guis.append(gui_addr)
                
        if local_guis:
            self.send_local_data(local_guis)
        if distant_guis:
            self.send_distant_data(distant_guis)

    def client_name_and_uuid(self, client_name: str, uuid: int):
        self.send_gui('/ray/gui/patchbay/client_name_and_uuid',
                      client_name, uuid)

    def port_added(self, pname: str, ptype: int, pflags: int, puuid: int):
        self.send_gui('/ray/gui/patchbay/port_added',
                      pname, ptype, pflags, puuid) 

    def port_renamed(self, ex_name: str, new_name):
        self.send_gui('/ray/gui/patchbay/port_renamed', ex_name, new_name)
    
    def port_removed(self, port_name: str):
        self.send_gui('/ray/gui/patchbay/port_removed', port_name)
    
    def metadata_updated(self, uuid: int, key: str, value: str):
        self.send_gui('/ray/gui/patchbay/metadata_updated', uuid, key, value)
    
    def connection_added(self, connection: tuple[str, str]):
        self.send_gui('/ray/gui/patchbay/connection_added',
                     connection[0], connection[1])

    def connection_removed(self, connection: tuple[str, str]):
        self.send_gui('/ray/gui/patchbay/connection_removed',
                     connection[0], connection[1])
    
    def server_stopped(self):
        # here server is JACK (in future maybe pipewire)
        self.send_gui('/ray/gui/patchbay/server_stopped')
    
    def send_transport_position(self, tpos: 'TransportPosition'):
        self.send_gui('/ray/gui/patchbay/transport_position',
                      tpos.frame, int(tpos.rolling), int(tpos.valid_bbt),
                      tpos.bar, tpos.beat, tpos.tick, tpos.beats_per_minutes)
    
    def send_dsp_load(self, dsp_load: int):
        self.send_gui('/ray/gui/patchbay/dsp_load', dsp_load)
    
    def send_one_xrun(self):
        self.send_gui('/ray/gui/patchbay/add_xrun')
    
    def send_buffersize(self):
        self.send_gui('/ray/gui/patchbay/buffer_size',
                     self.main_object.buffer_size)
    
    def send_samplerate(self):
        self.send_gui('/ray/gui/patchbay/sample_rate',
                     self.main_object.samplerate)
    
    def is_terminate(self):
        return self._terminate
    
    def send_server_lose(self):
        self.send_gui('/ray/gui/patchbay/server_lose')
        
        # In the case server is not responding
        # and gui has not yet been added to gui_list
        # but gui url stocked in self._tmp_gui_url
        if not self.gui_list and self._tmp_gui_url:
            try:
                addr = Address(self._tmp_gui_url)
            except:
                return
        
        self.send(addr, '/ray/gui/patchbay/server_lose')




class PortData:
    name: str
    type: int
    flags: int
    uuid: int
    
    def __init__(self, name: str, type: int, flags: int, uuid: int):
        self.name = name
        self.type = type
        self.flags = flags
        self.uuid = uuid


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# PatchBay Canvas engine using QGraphicsView/Scene
# Copyright (C) 2010-2019 Filipe Coelho <falktx@falktx.com>
# Copyright (C) 2019-2022 Mathieu Picot <picotmathieu@gmail.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of
# the License, or any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# For a full copy of the GNU General Public License see the doc/GPL.txt file.


# Imports (Globals)
import logging
from math import floor
import math
from shutil import move
import time
from typing import Iterator

from PyQt5.QtCore import (QT_VERSION, pyqtSignal, pyqtSlot,
                          Qt, QPoint, QPointF, QRectF, QTimer, QMarginsF)
from PyQt5.QtGui import QCursor, QPixmap, QPolygonF, QBrush
from PyQt5.QtWidgets import (QGraphicsRectItem, QGraphicsScene, QApplication,
                             QGraphicsView, QGraphicsItem)


# Imports (locals)
from .init_values import (
    CanvasItemType,
    PortMode,
    Direction,
    canvas,
    options,
    CallbackAct,
    MAX_PLUGIN_ID_ALLOWED)

from .box_widget import BoxWidget
from .connectable_widget import ConnectableWidget
from .line_widget import LineWidget
from .icon_widget import IconPixmapWidget, IconSvgWidget


_logger = logging.getLogger(__name__)


class RubberbandRect(QGraphicsRectItem):
    " This class is used by rectangle selection when user "
    " press mouse button and move to select boxes. "
    def __init__(self, scene: QGraphicsScene):
        QGraphicsRectItem.__init__(self, QRectF(0, 0, 0, 0))

        self.setZValue(-1)
        self.hide()

        scene.addItem(self)

    def type(self) -> CanvasItemType:
        return CanvasItemType.RUBBERBAND


class MovingBox:
    widget: BoxWidget
    from_pt: QPointF
    to_pt: QPoint
    start_time: float
    joining: bool


class WrappingBox:
    widget: BoxWidget
    wrap: bool


class PatchSceneMoth(QGraphicsScene):
    " This class is used for the scene. "
    " The child class in scene.py has all things to manage"
    " repulsives boxes."
    scale_changed = pyqtSignal(float)
    scene_group_moved = pyqtSignal(int, int, QPointF)
    plugin_selected = pyqtSignal(list)

    def __init__(self, view: QGraphicsView):
        QGraphicsScene.__init__(self)

        self._scale_area = False
        self._mouse_down_init = False
        self._mouse_rubberband = False
        self._mid_button_down = False
        self._pointer_border = QRectF(0.0, 0.0, 1.0, 1.0)
        self._scale_min = 0.1
        self._scale_max = 4.0

        self._rubberband = RubberbandRect(self)
        self._rubberband_selection = False
        self._rubberband_orig_point = QPointF(0, 0)

        self._view = view
        if not self._view:
            _logger.critical("Invalid view")
            return

        self._cursor_cut = None
        self._cursor_zoom_area = None

        self.move_boxes = list[MovingBox]()
        self.wrapping_boxes = list[WrappingBox]()
        self._MOVE_DURATION = 0.300 # 300ms
        self._MOVE_TIMER_INTERVAL = 20 # 20 ms step animation (50 Hz)
        self._move_timer_start_at = 0
        self._move_box_timer = QTimer()
        self._move_box_timer.setInterval(self._MOVE_TIMER_INTERVAL)
        self._move_box_timer.timeout.connect(self.move_boxes_animation)

        self.resizing_scene = False
        self.translating_view = False
        self._last_border_translate = QPointF(0.0, 0.0)

        self._borders_nav_timer = QTimer()
        self._borders_nav_timer.setInterval(50)
        self._borders_nav_timer.timeout.connect(self._cursor_view_navigation)
        self._last_view_cpos = QPointF()
        self._allowed_nav_directions = set[Direction]()

        self.flying_connectable = None

        self.selectionChanged.connect(self._slot_selection_changed)

    def deplace_boxes_from_repulsers(self, repulser_boxes: list[BoxWidget],
                                     wanted_direction=Direction.NONE,
                                     new_scene_rect=None):
        ''' This function change the place of boxes in order to have no 
            box overlapping other boxes.'''
        # just for easier syntax, this method is overloaded in scene.py
        # but executed in this file too.
        pass

    def clear(self):
        # reimplement Qt function and fix missing rubberband after clear
        QGraphicsScene.clear(self)
        self._rubberband = RubberbandRect(self)
        self.update_theme()

    def screen_position(self, point: QPointF) -> QPoint:
        return self._view.mapToGlobal(self._view.mapFromScene(point))

    def get_device_pixel_ratio_f(self):
        if QT_VERSION < 0x50600:
            return 1.0

        return self._view.devicePixelRatioF()

    def get_scale_factor(self):
        return self._view.transform().m11()

    def fix_scale_factor(self, transform=None):
        fix, set_view = False, False
        if not transform:
            set_view = True
            view = self._view
            transform = view.transform()

        scale = transform.m11()
        if scale > self._scale_max:
            fix = True
            transform.reset()
            transform.scale(self._scale_max, self._scale_max)
        elif scale < self._scale_min:
            fix = True
            transform.reset()
            transform.scale(self._scale_min, self._scale_min)

        if set_view:
            if fix:
                view.setTransform(transform)
            self.scale_changed.emit(transform.m11())

        return fix

    def _cursor_view_navigation(self):
        ''' This function is called every 50 ms when mouse
            left button is pressed. It moves the view if the mouse cursor
            is near a border of the view (in the limits of the scene size). '''
        # max speed of the drag when mouse is at a side pixel of the view  
        SPEED = 0.8
        
        # Acceleration, doesn't affect max speed
        POWER = 14

        view_width = self._view.width()
        view_height = self._view.height()
        if self._view.verticalScrollBar().isVisible():
            view_width -= self._view.verticalScrollBar().width()
        if self._view.horizontalScrollBar().isVisible():
            view_height -= self._view.horizontalScrollBar().height()
        
        view_cpos = self._view.mapFromGlobal(QCursor.pos())
        scene_cpos = self._view.mapToScene(view_cpos)
        
        # The scene relative area we want to be visible in the view 
        ensure_rect = QRectF(scene_cpos.x() - 1.0,
                             scene_cpos.y() - 1.0,
                             2.0, 2.0)
        
        # the scene relative area currently visible in the view
        vs_rect = QRectF(
            QPointF(self._view.mapToScene(0, 0)),
            QPointF(self._view.mapToScene(view_width - 1, view_height -1)))
        
        # The speed of the move depends on the scene size
        # to allow fast moves from one scene corner to another one.
        speed_hor = (SPEED * (self.sceneRect().width() - vs_rect.width())
                     / vs_rect.width()) 
        speed_ver = (SPEED * (self.sceneRect().height() - vs_rect.height())
                     / vs_rect.height())

        interval_hor = vs_rect.width() / 2
        interval_ver = vs_rect.height() / 2
        
        # Navigation is allowed in a direction only if mouse cursor has
        # already been moved in this direction, in order to prevent unintended
        # moves when user just pressed the mouse button.
        if self._last_view_cpos.isNull():
            self._allowed_nav_directions.clear()
        elif len(self._allowed_nav_directions) < 4:
            if view_cpos.x() < self._last_view_cpos.x():
                self._allowed_nav_directions.add(Direction.LEFT)
            elif view_cpos.x() > self._last_view_cpos.x():
                self._allowed_nav_directions.add(Direction.RIGHT)

            if view_cpos.y() < self._last_view_cpos.y():
                self._allowed_nav_directions.add(Direction.UP)
            elif view_cpos.y() > self._last_view_cpos.y():
                self._allowed_nav_directions.add(Direction.DOWN)

        self._last_view_cpos = view_cpos
        
        # Define the limits we want to see in the view
        # Note that the zone where there is no move is defined by the fact
        # the move in a direction is converted from float to int.
        # This way, a move lower than 1.0 pixel will be ignored.
        # By chance, the lower possible speed is good
        # 1.0 pixel * (1s / 0.050s) = 20 pixels/second. 
        
        apply = False
        
        if (Direction.LEFT in self._allowed_nav_directions
                and scene_cpos.x() < vs_rect.center().x()):
            offset = vs_rect.center().x() - max(scene_cpos.x(), vs_rect.left())
            ratio_x = offset / interval_hor
            move_x = - speed_hor * ((ratio_x ** POWER) * interval_hor)
            left = vs_rect.left() + int(move_x)
            ensure_rect.moveLeft(max(left, self.sceneRect().left()))
            if int(move_x):
                apply = True

        elif (Direction.RIGHT in self._allowed_nav_directions
                and scene_cpos.x() > vs_rect.center().x()):
            offset = min(scene_cpos.x(), vs_rect.right()) - vs_rect.center().x()
            ratio_x = offset / interval_hor
            move_x = speed_hor * ((ratio_x ** POWER) * interval_hor)
            right = vs_rect.right() + int(move_x)
            ensure_rect.moveRight(min(right, self.sceneRect().right()))
            if int(move_x):
                apply = True
            
        if (Direction.UP in self._allowed_nav_directions
                and scene_cpos.y() < vs_rect.center().y()):
            offset = vs_rect.center().y() - max(scene_cpos.y(), vs_rect.top())
            ratio_y = offset / interval_ver
            move_y = - speed_ver * ((ratio_y ** POWER) * interval_ver)
            top = vs_rect.top() + int(move_y)
            ensure_rect.moveTop(max(top, self.sceneRect().top()))
            if int(move_y):
                apply = True
        
        elif (Direction.DOWN in self._allowed_nav_directions
                and scene_cpos.y() > vs_rect.center().y()):
            offset = min(scene_cpos.y(), vs_rect.bottom()) - vs_rect.center().y()
            ratio_y = offset / interval_ver
            move_y = speed_ver * ((ratio_y ** POWER) * interval_ver)
            bottom = vs_rect.bottom() + int(move_y)
            ensure_rect.moveBottom(min(bottom, self.sceneRect().bottom()))
            if int(move_y):
                apply = True
        
        if apply:
            self._view.ensureVisible(ensure_rect, 0, 0)

    def _start_navigation_on_borders(self):
        if (options.borders_navigation
                and not self._borders_nav_timer.isActive()):
            self._last_view_cpos = QPointF()
            self._borders_nav_timer.start()

    def fix_temporary_scroll_bars(self):
        if self._view is None:
            return

        if self._view.horizontalScrollBar().isVisible():
            self._view.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOn)
        else:
            self._view.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)

        if self._view.verticalScrollBar().isVisible():
            self._view.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOn)
        else:
            self._view.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOff)

    def reset_scroll_bars(self):
        self._view.setHorizontalScrollBarPolicy(Qt.ScrollBarAsNeeded)
        self._view.setVerticalScrollBarPolicy(Qt.ScrollBarAsNeeded)

    def move_boxes_animation(self):
        # Animation is nice but not the priority.
        # Do not ensure all steps are played
        # but just move the box where it has to go now
        time_since_start = time.time() - self._move_timer_start_at
        ratio = min(1.0, time_since_start / self._MOVE_DURATION)

        for moving_box in self.move_boxes:
            if moving_box.widget is not None:
                x = (moving_box.from_pt.x()
                     + ((moving_box.to_pt.x() - moving_box.from_pt.x())
                        * (ratio ** 0.6)))
                
                y = (moving_box.from_pt.y()
                     + ((moving_box.to_pt.y() - moving_box.from_pt.y())
                        * (ratio ** 0.6)))

                moving_box.widget.setPos(x, y)
                moving_box.widget.repaint_lines(fast_move=True)

        for wrapping_box in self.wrapping_boxes:
            if wrapping_box.widget is not None:
                if time_since_start >= self._MOVE_DURATION:
                    wrapping_box.widget.animate_wrapping(1.00)
                else:
                    wrapping_box.widget.animate_wrapping(ratio)
                
        self.resize_the_scene()

        if time_since_start >= self._MOVE_DURATION:
            # Animation is finished
            self._move_box_timer.stop()
            
            # box update positions is forbidden while widget is in self.move_boxes
            # So we copy the list before to clear it
            # then we can ask update_positions on widgets
            boxes = [mb.widget for mb in self.move_boxes if not mb.joining]
            self.move_boxes.clear()
            self.wrapping_boxes.clear()

            for box in boxes:
                if box is not None:
                    if box.update_positions_pending:
                        box.update_positions()
                    box.send_move_callback()

                    # self.deplace_boxes_from_repulsers([box])
            self.deplace_boxes_from_repulsers(boxes)
            canvas.qobject.move_boxes_finished.emit()

    def add_box_to_animation(self, box_widget: BoxWidget, to_x: int, to_y: int,
                             force_anim=True, joining=False):
        for moving_box in self.move_boxes:
            if moving_box.widget is box_widget:
                break
        else:
            if not force_anim:
                # if box is not in a current animation
                # and force_anim is False,
                # then box position is directly changed
                if box_widget is not None:
                    box_widget.setPos(int(to_x), int(to_y))
                return

            moving_box = MovingBox()
            moving_box.widget = box_widget
            self.move_boxes.append(moving_box)

        moving_box.from_pt = box_widget.pos()
        moving_box.to_pt = QPoint(to_x, to_y)
        moving_box.start_time = time.time() - self._move_timer_start_at
        moving_box.joining = joining

        if not self._move_box_timer.isActive():
            moving_box.start_time = 0.0
            self._move_timer_start_at = time.time()
            self._move_box_timer.start()

    def add_box_to_animation_wrapping(self, box_widget: BoxWidget, wrap: bool):
        for wrapping_box in self.wrapping_boxes:
            if wrapping_box.widget is box_widget:
                wrapping_box.wrap = wrap
                break
        else:
            wrapping_box = WrappingBox()
            wrapping_box.widget = box_widget
            wrapping_box.wrap = wrap
            self.wrapping_boxes.append(wrapping_box)
        
        if not self._move_box_timer.isActive():
            self._move_timer_start_at = time.time()
            self._move_box_timer.start()

    def center_view_on(self, widget):
        self._view.centerOn(widget)

    def get_connectable_item_at(
            self, pos: QPointF, origin: ConnectableWidget) -> ConnectableWidget:
        for item in self.items(pos, Qt.ContainsItemShape, Qt.AscendingOrder):
            if isinstance(item, ConnectableWidget) and item is not origin:
                return item

    def get_box_at(self, pos: QPointF) -> BoxWidget:
        for item in self.items(pos, Qt.ContainsItemShape, Qt.AscendingOrder):
            if isinstance(item, BoxWidget):
                return item
    
    def get_selected_boxes(self) -> list[BoxWidget]:
        return [i for i in self.selectedItems() if isinstance(i, BoxWidget)]

    def list_selected_boxes(self) -> Iterator[BoxWidget]:
        for item in self.selectedItems():
            if isinstance(item, BoxWidget):
                yield item

    def removeItem(self, item: QGraphicsItem):
        for child_item in item.childItems():
            QGraphicsScene.removeItem(self, child_item)
        QGraphicsScene.removeItem(self, item)

    def update_limits(self):
        w0 = canvas.size_rect.width()
        h0 = canvas.size_rect.height()
        w1 = self._view.width()
        h1 = self._view.height()
        self._scale_min = w1/w0 if w0/h0 > w1/h1 else h1/h0

    def update_theme(self):
        if canvas.theme.scene_background_image is not None:
            bg_brush = QBrush()
            bg_brush.setTextureImage(canvas.theme.scene_background_image)
            self.setBackgroundBrush(bg_brush)
        else:
            self.setBackgroundBrush(canvas.theme.scene_background_color)
        
        self._rubberband.setPen(canvas.theme.rubberband.fill_pen())
        self._rubberband.setBrush(canvas.theme.rubberband.background_color())

        cur_color = ("black" if canvas.theme.scene_background_color.blackF() < 0.5
                     else "white")
        self._cursor_cut = QCursor(QPixmap(f":/cursors/cut-{cur_color}.png"), 1, 1)
        self._cursor_zoom_area = QCursor(
            QPixmap(f":/cursors/zoom-area-{cur_color}.png"), 8, 7)

        for connection in canvas.list_connections():
            if connection.widget is not None:
                connection.widget.update_theme()

    def drawBackground(self, painter, rect):
        painter.save()
        painter.setPen(Qt.NoPen)
        
        if not canvas.theme.scene_background_image.isNull():
            canvas.theme.scene_background_image.setDevicePixelRatio(3.0)
            bg_brush = QBrush()
            bg_brush.setTextureImage(canvas.theme.scene_background_image)
            painter.setBrush(bg_brush)
            painter.drawRect(rect)

        painter.setBrush(canvas.theme.scene_background_color)        
        painter.drawRect(rect)
        painter.restore()

    def get_new_scene_rect(self) -> QRectF:
        full_rect = QRectF()

        for widget in canvas.list_boxes():
            if widget.get_current_port_mode() is PortMode.NULL:
                continue
            full_rect |= widget.sceneBoundingRect().marginsAdded(
                QMarginsF(50.0, 20.0, 50.0, 20.0))

        return full_rect

    def resize_the_scene(self):
        if not options.elastic:
            return

        scene_rect = self.get_new_scene_rect()
        
        if not scene_rect.isNull():
            self.resizing_scene = True
            self.setSceneRect(scene_rect)
            self.resizing_scene = False

    def set_elastic(self, yesno: bool):
        options.elastic = True
        self.resize_the_scene()
        options.elastic = yesno

        if not yesno:
            # resize the scene to a null QRectF to auto set sceneRect
            # always growing with items
            self.setSceneRect(QRectF())

            # add a fake item with the current canvas scene size
            # (calculated with items), and remove it.
            fake_item = QGraphicsRectItem(self.get_new_scene_rect())
            self.addItem(fake_item)
            self.update()
            self.removeItem(fake_item)

    def set_cursor(self, cursor: QCursor):
        if self._view is None:
            return
        
        self._view.viewport().setCursor(cursor)

    def unset_cursor(self):
        if self._view is None:
            return
        
        self._view.viewport().unsetCursor()

    def zoom_ratio(self, percent: float, force=False):
        ratio = percent / 100.0
        transform = self._view.transform()
        
        if not force and ratio == transform.m11():
            return
        
        transform.reset()
        transform.scale(ratio, ratio)
        self._view.setTransform(transform)

        for box in canvas.list_boxes():
            if box.top_icon:
                box.top_icon.update_zoom(ratio)
        
        self.scale_changed.emit(transform.m11())

    def zoom_fit(self):
        if self._view is None:
            return

        full_rect = QRectF()
        
        for item in self.items():
            if isinstance(item, BoxWidget) and item.isVisible():
                rect = item.sceneBoundingRect()
                
                if full_rect.isNull():
                    full_rect = rect
                    continue
                
                full_rect.setLeft(min(full_rect.left(), rect.left()))
                full_rect.setRight(max(full_rect.right(), rect.right()))
                full_rect.setTop(min(full_rect.top(), rect.top()))
                full_rect.setBottom(max(full_rect.bottom(), rect.bottom()))
                
        if full_rect.isNull():
            return
        
        self._view.fitInView(full_rect, Qt.KeepAspectRatio)
        self.fix_scale_factor()
        self.scale_changed.emit(self._view.transform().m11())

    def zoom_in(self):
        view = self._view
        transform = view.transform()
        if transform.m11() < self._scale_max:
            transform.scale(1.2, 1.2)
            if transform.m11() > self._scale_max:
                transform.reset()
                transform.scale(self._scale_max, self._scale_max)
            view.setTransform(transform)
        self.scale_changed.emit(transform.m11())

    def zoom_out(self):
        view = self._view
        transform = view.transform()
        if transform.m11() > self._scale_min:
            transform.scale(0.833333333333333, 0.833333333333333)
            if transform.m11() < self._scale_min:
                transform.reset()
                transform.scale(self._scale_min, self._scale_min)
            view.setTransform(transform)
        self.scale_changed.emit(transform.m11())

    def zoom_reset(self):
        transform = self._view.transform()
        transform.reset()

        default_scale = options.default_zoom / 100
        transform.scale(default_scale, default_scale)
        self._view.setTransform(transform)
        self.scale_changed.emit(default_scale)

    @pyqtSlot()
    def _slot_selection_changed(self):
        items_list = self.selectedItems()

        if len(items_list) == 0:
            self.plugin_selected.emit([])
            return

        plugin_list = []

        for item in items_list:
            if item and item.isVisible():
                group_item = None

                if isinstance(item, BoxWidget):
                    group_item = item
                elif isinstance(item, ConnectableWidget):
                    group_item = item.parentItem()

                if group_item is not None and group_item._plugin_id >= 0:
                    plugin_id = group_item._plugin_id
                    if plugin_id > MAX_PLUGIN_ID_ALLOWED:
                        plugin_id = 0
                    plugin_list.append(plugin_id)

        self.plugin_selected.emit(plugin_list)

    def _trigger_rubberband_scale(self):
        # TODO, should enable an auto-zoom on 
        # Ctrl+Right clic + drag (rubberband)
        self._scale_area = True

        if self._cursor_zoom_area:
            self.set_cursor(self._cursor_zoom_area)

    def get_zoom_scale(self):
        return self._view.transform().m11()

    def keyPressEvent(self, event):
        if not self._view:
            event.ignore()
            return

        if event.key() == Qt.Key_Control:
            if self._mid_button_down:
                self._start_connection_cut()

        elif event.key() == Qt.Key_Home:
            event.accept()
            self.zoom_fit()
            return

        elif QApplication.keyboardModifiers() & Qt.ControlModifier:
            if event.key() == Qt.Key_Plus:
                event.accept()
                self.zoom_in()
                return

            if event.key() == Qt.Key_Minus:
                event.accept()
                self.zoom_out()
                return

            if event.key() == Qt.Key_1:
                event.accept()
                self.zoom_reset()
                return

        QGraphicsScene.keyPressEvent(self, event)

    def keyReleaseEvent(self, event):
        if event.key() == Qt.Key_Control:
            # Connection cut mode off
            if self._mid_button_down:
                self.unset_cursor()

        QGraphicsScene.keyReleaseEvent(self, event)

    def _start_connection_cut(self):
        if self._cursor_cut:
            self.set_cursor(self._cursor_cut)

    def zoom_wheel(self, delta: int):
        transform = self._view.transform()
        scale = transform.m11()

        if ((delta > 0 and scale < self._scale_max)
                or (delta < 0 and scale > self._scale_min)):
            # prevent too large unzoom
            if delta < 0:
                rect = self.sceneRect()

                top_left_vw = self._view.mapFromScene(rect.topLeft())
                if (top_left_vw.x() > self._view.width() / 4
                        and top_left_vw.y() > self._view.height() / 4):
                    return

            # Apply scale
            factor = 1.4142135623730951 ** (delta / 240.0)
            transform.scale(factor, factor)
            self.fix_scale_factor(transform)
            self._view.setTransform(transform)
            self.scale_changed.emit(transform.m11())

            # Update box icons especially when they are not scalable
            # eg. coming from system theme
            for box in canvas.list_boxes():
                if box.top_icon:
                    box.top_icon.update_zoom(scale * factor)

    def mouseDoubleClickEvent(self, event):
        if event.button() == Qt.LeftButton and not canvas.menu_shown:
            # parse items under mouse to prevent CallbackAct.DOUBLE_CLICK
            # if mouse is on a box
            
            has_box = False
            items = self.items(
                event.scenePos(), Qt.ContainsItemShape, Qt.AscendingOrder)

            for item in items:
                if isinstance(item, ConnectableWidget):
                    # start a flying connection with mouse button not pressed
                    # here we just change the cursor and define the ConnectableWidget
                    self.flying_connectable = item
                    self.set_cursor(QCursor(Qt.CrossCursor))
                    self._start_navigation_on_borders()
                    return

                if not has_box:
                    has_box = isinstance(item, BoxWidget)
            
            if not has_box:
                canvas.callback(CallbackAct.BG_DOUBLE_CLICK)

        QGraphicsScene.mouseDoubleClickEvent(self, event)

    def mousePressEvent(self, event):
        if self.flying_connectable:
            if event.button() == Qt.LeftButton:
                self.flying_connectable.mouseReleaseEvent(event)
                self.flying_connectable = None
                self.set_cursor(Qt.ArrowCursor)
                return

            if event.button() == Qt.RightButton:
                self.flying_connectable.mousePressEvent(event)
                return

        ctrl_pressed = bool(QApplication.keyboardModifiers() & Qt.ControlModifier)
        self._mouse_down_init = bool(
            (event.button() == Qt.LeftButton and not ctrl_pressed)
            or (event.button() == Qt.RightButton and ctrl_pressed))
        
        self._mouse_rubberband = False

        if event.button() == Qt.MidButton:
            if ctrl_pressed:
                self._mid_button_down = True
                self._start_connection_cut()

                pos = event.scenePos()
                self._pointer_border.moveTo(floor(pos.x()), floor(pos.y()))

                for item in self.items(self._pointer_border):
                    if isinstance(item, (ConnectableWidget, LineWidget)):
                        item.trigger_disconnect()

        QGraphicsScene.mousePressEvent(self, event)
        canvas.menu_shown = False

        if (self._view.dragMode() != QGraphicsView.ScrollHandDrag
                and event.buttons() == Qt.LeftButton):
            # Middle click and move changes the view drag mode
            # and fake a left button.
            # we don't want navigation on borders in case of middle click.
            self._start_navigation_on_borders()

    def mouseMoveEvent(self, event):
        if self.flying_connectable is not None:
            self.flying_connectable.mouseMoveEvent(event)
            return
        
        if self._mouse_down_init:
            self._mouse_down_init = False
            topmost = self.itemAt(event.scenePos(), self._view.transform())
            self._mouse_rubberband = (
                not isinstance(topmost, (BoxWidget, ConnectableWidget,
                                         IconPixmapWidget, IconSvgWidget))
                and int(event.buttons()))

        if self._mouse_rubberband:
            event.accept()
            pos = event.scenePos()
            pos_x = pos.x()
            pos_y = pos.y()
            if not self._rubberband_selection:
                self._rubberband.show()
                self._rubberband_selection = True
                self._rubberband_orig_point = pos
            rubb_orig_point = self._rubberband_orig_point

            x = min(pos_x, rubb_orig_point.x())
            y = min(pos_y, rubb_orig_point.y())

            line_hinting = canvas.theme.rubberband.fill_pen().widthF() / 2.0
            self._rubberband.setRect(
                x + line_hinting, y + line_hinting,
                abs(pos_x - rubb_orig_point.x()),
                abs(pos_y - rubb_orig_point.y()))

        if (self._mid_button_down
                and QApplication.keyboardModifiers() & Qt.ControlModifier):
            for item in self.items(
                    QPolygonF([event.scenePos(), event.lastScenePos(),
                               event.scenePos()])):
                if isinstance(item, LineWidget):
                    item.trigger_disconnect()
        
        QGraphicsScene.mouseMoveEvent(self, event)

    def mouseReleaseEvent(self, event):
        if self.flying_connectable:
            QGraphicsScene.mouseReleaseEvent(self, event)
            return
        
        if self._scale_area and not self._rubberband_selection:
            self._scale_area = False
            self.unset_cursor()

        if self._rubberband_selection:
            if self._scale_area:
                self._scale_area = False
                self.unset_cursor()

                rect = self._rubberband.rect()
                self._view.fitInView(
                    rect.x(), rect.y(),
                    rect.width(), rect.height(), Qt.KeepAspectRatio)
                self.fix_scale_factor()

            else:
                for item in self.items():
                    if isinstance(item, BoxWidget):
                        item_rect = item.sceneBoundingRect()
                        if self._rubberband.rect().contains(item_rect):
                            item.setSelected(True)

            self._rubberband.hide()
            self._rubberband.setRect(0, 0, 0, 0)
            self._rubberband_selection = False

        else:
            for item in self.get_selected_boxes():
                item.check_item_pos()
                self.scene_group_moved.emit(
                    item.get_group_id(), item.get_splitted_mode(),
                    item.scenePos())

            if len(self.selectedItems()) > 1:
                self.update()

        self._mouse_down_init = False
        self._mouse_rubberband = False

        if event.button() == Qt.LeftButton:
            self._borders_nav_timer.stop()

        if event.button() == Qt.MidButton:
            event.accept()

            self._mid_button_down = False

            # Connection cut mode off
            if QApplication.keyboardModifiers() & Qt.ControlModifier:
                self.unset_cursor()
            return

        QGraphicsScene.mouseReleaseEvent(self, event)

    def wheelEvent(self, event):
        if not self._view:
            event.ignore()
            return

        if QApplication.keyboardModifiers() & Qt.ControlModifier:
            self.zoom_wheel(event.delta())
            event.accept()
            return

        QGraphicsScene.wheelEvent(self, event)

    def contextMenuEvent(self, event):
        if canvas.is_line_mov:
            event.ignore()
            return
        
        for item in self.items(event.scenePos()):
            if isinstance(item, (BoxWidget, ConnectableWidget)):
                break
        else:
            if QApplication.keyboardModifiers() & Qt.ControlModifier:
                event.accept()
                self._trigger_rubberband_scale()
                return

            event.accept()
            sc_pos = event.screenPos()
            canvas.callback(CallbackAct.BG_RIGHT_CLICK, sc_pos.x(), sc_pos.y())
            return

        QGraphicsScene.contextMenuEvent(self, event)



#!/usr/bin/env python3
"""Setup file for pyjacklib."""

from os.path import dirname, join

from setuptools import setup


def read(*args):
    return open(join(dirname(__file__), *args)).read()


PKGDIR = 'jacklib'
version = {}
exec(read(PKGDIR, 'version.py'), {}, version)
setup(version=version["__version__"], packages=[PKGDIR])



import argparse
import os
from pathlib import Path
import sys
from typing import TYPE_CHECKING
from PyQt5.QtCore import QSettings, QSize, QFile, QObject, pyqtSignal
from PyQt5.QtWidgets import QApplication, QWidget
from PyQt5.QtGui import QIcon, QPixmap, QPalette
try:
    from liblo import Address
except ImportError:
    from pyliblo3 import Address

import ray

if TYPE_CHECKING:
    from gui_signaler import Signaler

_translate = QApplication.translate

_RAY_ICONS_CACHE_LIGHT = {}
_RAY_ICONS_CACHE_DARK = {}
_APP_ICONS_CACHE_LIGHT = {}
_APP_ICONS_CACHE_DARK = {}

class RS:
    settings = QSettings()

    # HD for Hideable dialog
    # we can add elements here but never change 
    # the existing ones, because they are used in user config files
    HD_Donations = 0x001
    HD_OpenNsmSession = 0x002
    HD_SnapshotsInfo = 0x004
    HD_WaitCloseUser = 0x008
    HD_JackConfigScript = 0x010
    HD_SessionScripts = 0x020
    HD_SystrayClose = 0x040
    HD_StartupRecentSessions = 0x080
    HD_ArdourConversion = 0x100
    
    _signaler: 'Signaler' = None
    
    @classmethod
    def set_settings(cls, settings):
        del cls.settings
        cls.settings = settings

    @classmethod
    def is_hidden(cls, hideable_dialog: int)->bool:
        hidden_dialogs = cls.settings.value('hidden_dialogs', 0, type=int)
        return bool(hidden_dialogs & hideable_dialog)

    @classmethod
    def set_hidden(cls, hiddeable_dialog: int, hide=True):
        hidden_dialogs = cls.settings.value('hidden_dialogs', 0, type=int)

        if hide:
            hidden_dialogs |= hiddeable_dialog
        else:
            hidden_dialogs &= ~hiddeable_dialog

        cls.settings.setValue('hidden_dialogs', hidden_dialogs)
        
        if cls._signaler is not None:
            cls._signaler.hiddens_changed.emit(hidden_dialogs)

    @classmethod
    def reset_hiddens(cls):
        cls.settings.setValue('hidden_dialogs', 0)
        if cls._signaler is not None:
            cls._signaler.hiddens_changed.emit(0)
        
    @classmethod
    def set_signaler(cls, signaler: 'Signaler'):
        cls._signaler = signaler


class ErrDaemon:
    # for use on network session under NSM
    NO_ERROR = 0
    NO_ANNOUNCE = -1
    NOT_OFF = -2
    WRONG_ROOT = -3
    FORBIDDEN_ROOT = -4
    NOT_NSM_LOCKED = -5
    WRONG_VERSION = -6


class RayAbstractIcon(QIcon):
    def __init__(self, icon_name: str, dark=False):
        QIcon.__init__(self)
        breeze = 'breeze-dark' if dark else 'breeze'
        self.addFile(':scalable/%s/%s' % (breeze, icon_name), QSize(22, 22))
        self.addPixmap(
            QPixmap(
                ':scalable/%s/disabled/%s' %
                (breeze, icon_name)), QIcon.Disabled, QIcon.Off)


def RayIcon(icon_name: str, dark=False) -> RayAbstractIcon:
    if dark and icon_name in _RAY_ICONS_CACHE_DARK.keys():
        return _RAY_ICONS_CACHE_DARK[icon_name]
    if not dark and icon_name in _RAY_ICONS_CACHE_LIGHT.keys():
        return _RAY_ICONS_CACHE_LIGHT[icon_name]
    
    icon = RayAbstractIcon(icon_name, dark)
    if dark:
        _RAY_ICONS_CACHE_DARK[icon_name] = icon
    else:
        _RAY_ICONS_CACHE_LIGHT[icon_name] = icon
    
    return icon


class CommandLineArgs(argparse.Namespace):
    daemon_url: Address = None
    daemon_port: Address = None
    out_daemon = False
    session_root: str = ''
    config_dir = ''
    debug = False
    debug_only = False
    no_client_messages = False
    net_session_root = ''
    net_daemon_id = 0
    under_nsm = False
    NSM_URL = ''
    session_root = ''
    start_session = ''
    force_new_daemon = False

    @classmethod
    def eat_attributes(cls, parsed_args):
        for attr_name in dir(parsed_args):
            if not attr_name.startswith('_'):
                setattr(cls, attr_name, getattr(parsed_args, attr_name))

        if cls.debug_only:
            cls.debug = True

        if cls.debug or cls.no_client_messages:
            cls.force_new_daemon = True

        if cls.config_dir and not os.access(cls.config_dir, os.W_OK):
            sys.stderr.write(
                '%s is not a writable config dir, try another one\n'
                % cls.config_dir)
            sys.exit(1)

        if os.getenv('NSM_URL'):
            try:
                cls.NSM_URL = ray.get_liblo_address(os.getenv('NSM_URL'))
            except BaseException:
                sys.stderr.write('%s is not a valid NSM_URL\n'
                                 % os.getenv('NSM_URL'))
                sys.exit(1)

            cls.under_nsm = True

        if (cls.session_root is not None
                and cls.session_root.endswith('/')):
            cls.session_root = cls.session_root[:-1]

    @classmethod
    def change_session_root(cls, path: str):
        cls.session_root = path


class ArgParser(argparse.ArgumentParser):
    def __init__(self):
        argparse.ArgumentParser.__init__(
            self,
            description=_translate(
                'help',
                'A session manager based on the Non-Session-Manager API '
                + 'for sound applications.'))
        self.add_argument('--daemon-url', '-u', type=ray.get_liblo_address,
                          help=_translate('help',
                                          'connect to this daemon url'))
        self.add_argument('--daemon-port', '-p',
                          type=ray.get_liblo_address_from_port,
                          help=_translate('help',
                                          'connect to this daemon port'))
        self.add_argument('--out-daemon', action='store_true',
                          help=argparse.SUPPRESS)
        self.add_argument('--session-root', '-r', type=str,
                          help=_translate(
                              'help', 'Use this folder as root for sessions'))
        self.add_argument('--start-session', '-s', type=str,
                          help=_translate('help',
                                          'Open this session at startup'))
        self.add_argument('--config-dir', '-c', type=str, default='',
                          help=_translate('help', 'use a custom config dir'))
        self.add_argument('--debug', '-d', action='store_true',
                          help=_translate('help', 'display OSC messages'))
        self.add_argument('--debug-only', '-do', action='store_true',
                          help=_translate('help',
                                          'debug without client messages'))
        self.add_argument('---no-client-messages', '-ncm', action='store_true',
                          help=_translate('help',
                                          'do not print client messages'))
        self.add_argument(
            '--force-new-daemon', '-fnd', action='store_true',
            help=_translate('help', 'prevent to attach to an already running daemon'))
        self.add_argument('--net-session-root', type=str, default='',
                          help=argparse.SUPPRESS)
        self.add_argument('--net-daemon-id', type=int, default=0,
                          help=argparse.SUPPRESS)
        self.add_argument('-v', '--version', action='version',
                          version=ray.VERSION)

        parsed_args = argparse.ArgumentParser.parse_args(self)
        CommandLineArgs.eat_attributes(parsed_args)


def init_gui_tools():
    if CommandLineArgs.under_nsm:
        settings = QSettings('%s/child_sessions'
                             % QApplication.organizationName())
    elif CommandLineArgs.config_dir:
        settings = QSettings(CommandLineArgs.config_dir)
    else:
        settings = QSettings()

    RS.set_settings(settings)

    if not CommandLineArgs.session_root:
        CommandLineArgs.change_session_root(
            settings.value('default_session_root',
                           ray.DEFAULT_SESSION_ROOT,
                           type=str))

def is_dark_theme(widget: QWidget) -> bool:
    return bool(
        widget.palette().brush(QPalette.Active, QPalette.WindowText).color().lightness()
        > 128)

def split_in_two(string: str) -> tuple[str, str]:
        middle = int(len(string)/2)
        sep_indexes = []
        last_was_digit = False

        for sep in (' ', '-', '_', 'capital'):
            for i in range(len(string)):
                c = string[i]
                if sep == 'capital':
                    if c.upper() == c:
                        if not c.isdigit() or not last_was_digit:
                            sep_indexes.append(i)
                        last_was_digit = c.isdigit()

                elif c == sep:
                    sep_indexes.append(i)

            if sep_indexes:
                break

        if not sep_indexes or sep_indexes == [0]:
            return (string, '')

        best_index = 0
        best_dif = middle

        for s in sep_indexes:
            dif = abs(middle - s)
            if dif < best_dif:
                best_index = s
                best_dif = dif

        if sep == ' ':
            return (string[:best_index], string[best_index+1:])
        return (string[:best_index], string[best_index:])

def dirname(*args) -> str:
    return os.path.dirname(*args)

def basename(*args) -> str:
    return os.path.basename(*args)

def get_code_root() -> Path:
    return Path(__file__).parent.parent.parent

def server_status_string(server_status: int) -> str:
    server_status_strings = {
        ray.ServerStatus.OFF     : _translate('server status', "off"),
        ray.ServerStatus.NEW     : _translate('server status', "new"),
        ray.ServerStatus.OPEN    : _translate('server status', "open"),
        ray.ServerStatus.CLEAR   : _translate('server status', "clear"),
        ray.ServerStatus.SWITCH  : _translate('server status', "switch"),
        ray.ServerStatus.LAUNCH  : _translate('server status', "launch"),
        ray.ServerStatus.PRECOPY : _translate('server status', "copy"),
        ray.ServerStatus.COPY    : _translate('server status', "copy"),
        ray.ServerStatus.READY   : _translate('server status', "ready"),
        ray.ServerStatus.SAVE    : _translate('server status', "save"),
        ray.ServerStatus.CLOSE   : _translate('server status', "close"),
        ray.ServerStatus.SNAPSHOT: _translate('server_status', "snapshot"),
        ray.ServerStatus.REWIND  : _translate('server_status', "rewind"),
        ray.ServerStatus.WAIT_USER   : _translate('server_status', "waiting"),
        ray.ServerStatus.OUT_SAVE    : _translate('server_status', "save"),
        ray.ServerStatus.OUT_SNAPSHOT: _translate('server_status', "snapshot"),
        ray.ServerStatus.SCRIPT  : _translate('server_status', "script")}

    if not 0 <= server_status < len(server_status_strings):
        return _translate('server status', "invalid")

    return server_status_strings[server_status]

def client_status_string(client_status: int) -> str:
    client_status_strings = {
        ray.ClientStatus.STOPPED: _translate('client status', "stopped"),
        ray.ClientStatus.LAUNCH : _translate('client status', "launch"),
        ray.ClientStatus.OPEN   : _translate('client status', "open"),
        ray.ClientStatus.READY  : _translate('client status', "ready"),
        ray.ClientStatus.PRECOPY: _translate('client status', "copy"),
        ray.ClientStatus.COPY   : _translate('client status', "copy"),
        ray.ClientStatus.SAVE   : _translate('client status', "save"),
        ray.ClientStatus.SWITCH : _translate('client status', "switch"),
        ray.ClientStatus.QUIT   : _translate('client status', "quit"),
        ray.ClientStatus.NOOP   : _translate('client status', "noop"),
        ray.ClientStatus.ERROR  : _translate('client status', "error"),
        ray.ClientStatus.REMOVED: _translate('client status', "removed"),
        ray.ClientStatus.UNDEF  : _translate('client_status', ""),
        ray.ClientStatus.SCRIPT : _translate('client_status', 'script'),
        ray.ClientStatus.LOSE   : _translate('client_status', "lose")}

    if not 0 <= client_status < len(client_status_strings):
        return _translate('client_status', 'invalid')

    return client_status_strings[client_status]

def error_text(error:int)->str:
    text = ''
    
    if error == ray.Err.SESSION_IN_SESSION_DIR:
        text = _translate(
            'guimsg', 
            """Can't create session in a dir containing a session
for better organization.""")
    
    return text

def get_app_icon(icon_name: str, widget: QWidget) -> QIcon:
    dark = bool(
        widget.palette().brush(
            2, QPalette.WindowText).color().lightness() > 128)
    
    if dark and icon_name in _APP_ICONS_CACHE_DARK.keys():
        return _APP_ICONS_CACHE_DARK[icon_name]
    if not dark and icon_name in _APP_ICONS_CACHE_LIGHT.keys():
        return _APP_ICONS_CACHE_LIGHT[icon_name]
    
    icon = QIcon.fromTheme(icon_name)

    if icon.isNull():
        for ext in ('svg', 'svgz', 'png'):
            filename = ":app_icons/%s.%s" % (icon_name, ext)
            darkname = ":app_icons/dark/%s.%s" % (icon_name, ext)

            if dark and QFile.exists(darkname):
                filename = darkname

            if QFile.exists(filename):
                del icon
                icon = QIcon()
                icon.addFile(filename)
                break

    if icon.isNull():
        for path in ('/usr/local', '/usr', '%s/.local' % os.getenv('HOME')):
            for ext in ('png', 'svg', 'svgz', 'xpm'):
                filename = "%s/share/pixmaps/%s.%s" % (path, icon_name, ext)
                if QFile.exists(filename):
                    del icon
                    icon = QIcon()
                    icon.addFile(filename)
                    break

    if dark:
        _APP_ICONS_CACHE_DARK[icon_name] = icon
    else:
        _APP_ICONS_CACHE_LIGHT[icon_name] = icon

    return icon



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# PatchBay Canvas engine using QGraphicsView/Scene
# Copyright (C) 2010-2019 Filipe Coelho <falktx@falktx.com>
# Copyright (C) 2019-2022 Mathieu Picot <picotmathieu@gmail.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of
# the License, or any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# For a full copy of the GNU General Public License see the doc/GPL.txt file.

from dataclasses import dataclass
from PyQt5.QtCore import (QRectF, QMarginsF, Qt)
from PyQt5.QtWidgets import QGraphicsView

from .init_values import (
    canvas,
    options,
    PortMode,
    Direction)

from .scene_moth import PatchSceneMoth
from .box_widget import BoxWidget


@dataclass
class BoxAndRect:
    rect: QRectF
    item: BoxWidget
    
        
@dataclass
class ToMoveBox:
    directions: list[Direction]
    item: BoxWidget
    rect: QRectF
    repulser: BoxAndRect
        
    def __lt__(self, other: 'ToMoveBox'):
        if self.directions != other.directions:
            return self.directions < other.directions
        
        # should not happen
        if not self.directions:
            return True
        
        last_direc = self.directions[-1]
        if last_direc is Direction.LEFT:
            return self.rect.right() > other.rect.right()
        if last_direc is Direction.UP:
            return self.rect.bottom() > other.rect.bottom()
        if last_direc is Direction.RIGHT:
            return self.rect.left() < other.rect.left()
        if last_direc is Direction.DOWN:
            return self.rect.top() < other.rect.top()
        
        # should not happen
        return True
    

class PatchScene(PatchSceneMoth):
    " This class part of the scene is for repulsive boxes option"
    " because the algorythm is not simple and takes a lot of lines."
    " See scene_abstract.py for others scene methods."
    def __init__(self, view: QGraphicsView):
        PatchSceneMoth.__init__(self, view)
        self._full_repulse_boxes = set[BoxWidget]()

    def deplace_boxes_from_repulsers(self, repulser_boxes: list[BoxWidget],
                                     wanted_direction=Direction.NONE,
                                     new_scene_rect=None):
        ''' This function change the place of boxes in order to have no box overlapping
            other boxes.'''
        def get_direction(fixed_rect: QRectF, moving_rect: QRectF,
                          parent_directions=list[Direction]()) -> Direction:
            if (moving_rect.top() <= fixed_rect.center().y() <= moving_rect.bottom()
                    or fixed_rect.top() <= moving_rect.center().y() <= fixed_rect.bottom()):
                if (fixed_rect.right() < moving_rect.center().x()
                        and fixed_rect.center().x() < moving_rect.left()):
                    if Direction.LEFT in parent_directions:
                        return Direction.LEFT
                    return Direction.RIGHT
                
                if (fixed_rect.left() > moving_rect.center().x()
                        and fixed_rect.center().x() > moving_rect.right()):
                    if Direction.RIGHT in parent_directions:
                        return Direction.RIGHT
                    return Direction.LEFT
            
            if fixed_rect.center().y() <= moving_rect.center().y():
                if Direction.UP in parent_directions:
                    return Direction.UP
                return Direction.DOWN
            
            if Direction.DOWN in parent_directions:
                return Direction.DOWN
            return Direction.UP
        
        def repulse(direction: Direction, fixed, moving,
                    fixed_port_mode: int, moving_port_mode: int) -> QRectF:
            ''' returns a QRectF to be placed at side of fixed_rect
                where fixed_rect is an already determinated futur place
                for a box '''
                
            if isinstance(fixed, BoxWidget):
                fixed_rect = fixed.boundingRect().translated(fixed.pos())
            else:
                fixed_rect = fixed
            
            if isinstance(moving, BoxWidget):
                rect = moving.boundingRect().translated(moving.pos())
            else:
                rect = moving
            
            x = rect.left()
            y = rect.top()
            
            if direction in (Direction.LEFT, Direction.RIGHT):
                spacing = box_spacing

                if direction == Direction.LEFT:
                    if (fixed_port_mode & PortMode.INPUT
                            or moving_port_mode & PortMode.OUTPUT):
                        spacing = box_spacing_hor
                    x = fixed_rect.left() - spacing - rect.width()
                    if x < 0:
                        x -= 1.0
                    x = float(int(x))
                else:
                    if (fixed_port_mode & PortMode.OUTPUT
                            or moving_port_mode & PortMode.INPUT):
                        spacing = box_spacing_hor
                    x = fixed_rect.right() + spacing
                    if x < 0:
                        x -= 1.0
                    x = float(int(x + 0.99))

                top_diff = abs(fixed_rect.top() - rect.top())
                bottom_diff = abs(fixed_rect.bottom() - rect.bottom())

                if bottom_diff > top_diff and top_diff <= magnet:
                    y = fixed_rect.top()
                elif bottom_diff <= magnet:
                    y = fixed_rect.bottom() - rect.height()
            
            elif direction in (Direction.UP, Direction.DOWN):
                if direction == Direction.UP:
                    y = fixed_rect.top() - box_spacing - rect.height()
                    if y < 0:
                        y -= 1.0
                    y = float(int(y))
                else:
                    y = fixed_rect.bottom() + box_spacing
                    if y < 0:
                        y -= 1.0
                    y = float(int(y + 0.99))
                
                left_diff = abs(fixed_rect.left() - rect.left())
                right_diff = abs(fixed_rect.right() - rect.right())
                
                if right_diff > left_diff and left_diff <= magnet:
                    x = fixed_rect.left()
                elif right_diff <= magnet:
                    x = fixed_rect.right() - rect.width()

            return QRectF(x, y, rect.width(), rect.height())

        def rect_has_to_move_from(
                repulser_rect: QRectF, rect: QRectF,
                repulser_port_mode: int, rect_port_mode: int)->bool:
            left_spacing = right_spacing = box_spacing
            
            if (repulser_port_mode & PortMode.INPUT
                    or rect_port_mode & PortMode.OUTPUT):
                left_spacing = box_spacing_hor
            
            if (repulser_port_mode & PortMode.OUTPUT
                    or rect_port_mode & PortMode.INPUT):
                right_spacing = box_spacing_hor
            
            large_repulser_rect = repulser_rect.adjusted(
                - left_spacing, - box_spacing,
                right_spacing, box_spacing)

            return rect.intersects(large_repulser_rect)

        # --- function start ---
        if not options.prevent_overlap:
            return
        
        box_spacing = canvas.theme.box_spacing
        box_spacing_hor = canvas.theme.box_spacing_horizontal
        magnet = canvas.theme.magnet

        to_move_boxes = list[ToMoveBox]()
        repulsers = list[BoxAndRect]()
        wanted_directions = [wanted_direction]
        
        for box in repulser_boxes:
            self._full_repulse_boxes.add(box)
            srect = box.boundingRect()

            if new_scene_rect is not None:
                srect = new_scene_rect
            else:
                # if box is already moving, consider its end position
                for moving_box in self.move_boxes:
                    if moving_box.widget is box:
                        if moving_box.joining:
                            # if this box is joining, it will be removed soon
                            # so, it has not to be a repulser
                            return
                        srect.translate(moving_box.to_pt)
                        break
                else:
                    srect.translate(box.pos())

            repulser = BoxAndRect(srect, box)
            repulsers.append(repulser)
            items_to_move = list[BoxAndRect]()

            search_rect = srect.marginsAdded(
                QMarginsF(canvas.theme.box_spacing_horizontal,
                          canvas.theme.box_spacing, 
                          canvas.theme.box_spacing_horizontal,
                          canvas.theme.box_spacing))

            for widget in self.items(search_rect, Qt.IntersectsItemShape,
                                     Qt.AscendingOrder):
                if not isinstance(widget, BoxWidget):
                    continue
                
                if (widget in repulser_boxes
                        or widget in [b.item for b in to_move_boxes]
                        or widget in [b.widget for b in self.move_boxes]):
                    continue

                irect = widget.sceneBoundingRect()

                if rect_has_to_move_from(
                        repulser.rect, irect,
                        repulser.item.get_current_port_mode(),
                        widget.get_current_port_mode()):
                    items_to_move.append(BoxAndRect(irect, widget))

            for moving_box in self.move_boxes:
                if (moving_box.widget in repulser_boxes
                        or moving_box.joining
                        or moving_box.widget in [b.item for b in to_move_boxes]):
                    continue

                widget = moving_box.widget
                irect = widget.boundingRect()
                irect.translate(moving_box.to_pt)
                
                if rect_has_to_move_from(
                        repulser.rect, irect,
                        repulser.item.get_current_port_mode(),
                        widget.get_current_port_mode()):
                    items_to_move.append(BoxAndRect(irect, widget))
            
            for item_to_move in items_to_move:
                item = item_to_move.item
                irect = item_to_move.rect
                    
                # evaluate in which direction should go the box
                direction = get_direction(srect, irect, wanted_directions)
                to_move_boxes.append(ToMoveBox([direction], item, irect, repulser))

        to_move_boxes.sort()

        # !!! to_move_boxes list is dynamic
        # elements can be added to the list while iteration !!!
        for to_move_box in to_move_boxes:
            item, repulser = to_move_box.item, to_move_box.repulser
            irect = item.boundingRect().translated(item.pos())

            directions = to_move_box.directions.copy()
            new_direction = get_direction(repulser.rect, irect, directions)
            directions.append(new_direction)

            # calculate the new position of the box repulsed by its repulser
            new_rect = repulse(new_direction, repulser.rect, item,
                               repulser.item.get_current_port_mode(),
                               item.get_current_port_mode())
            
            active_repulsers = list[BoxAndRect]()
            
            # while there is a repulser rect at new box position
            # move the future box position
            while True:
                # list just here to prevent infinite loop
                # we save the repulsers that already have moved the rect
                for repulser in repulsers:
                    if rect_has_to_move_from(
                            repulser.rect, new_rect,
                            repulser.item.get_current_port_mode(),
                            item.get_current_port_mode()):

                        if repulser in active_repulsers:
                            continue
                        active_repulsers.append(repulser)
                        
                        new_direction = get_direction(
                            repulser.rect, new_rect, directions)
                        new_rect = repulse(
                            new_direction, repulser.rect, new_rect,
                            repulser.item._current_port_mode,
                            item._current_port_mode)
                        directions.append(new_direction)
                        break
                else:
                    break

            # Now we know where the box will be definitely positioned
            # So, this is now a repulser for other boxes
            repulser = BoxAndRect(new_rect, item)
            repulsers.append(repulser)
            self._full_repulse_boxes.add(item)
            
            # check which existing boxes exists at the new place of the box
            # and add them to this to_move_boxes iteration
            adding_list = list[ToMoveBox]()

            search_rect = new_rect.marginsAdded(
                QMarginsF(canvas.theme.box_spacing_horizontal,
                          canvas.theme.box_spacing, 
                          canvas.theme.box_spacing_horizontal,
                          canvas.theme.box_spacing))
            
            for widget in self.items(search_rect):
                if not isinstance(widget, BoxWidget):
                    continue
                if (widget in [r.item for r in repulsers]
                        or widget in [b.item for b in to_move_boxes]
                        or widget in [b.widget for b in self.move_boxes]):
                    continue
                
                mirect = widget.sceneBoundingRect()
                if rect_has_to_move_from(
                        new_rect, mirect,
                        to_move_box.item.get_current_port_mode(),
                        widget.get_current_port_mode()):
                    
                    adding_list.append(
                        ToMoveBox(directions, widget, mirect, repulser))                        
            
            for moving_box in self.move_boxes:
                mitem = moving_box.widget
                assert isinstance(mitem, BoxWidget)
                
                if (mitem in [r.item for r in repulsers]
                        or moving_box.joining
                        or mitem in [b.item for b in to_move_boxes]):
                    continue

                rect = mitem.boundingRect()
                rect.translate(moving_box.to_pt)
                
                if rect_has_to_move_from(
                        new_rect, rect,
                        to_move_box.item.get_current_port_mode(),
                        mitem.get_current_port_mode()):

                    adding_list.append(
                        ToMoveBox(directions, moving_box.widget, rect, repulser))

            adding_list.sort()

            for to_move_box in adding_list:
                to_move_boxes.append(to_move_box)

            # now we decide where the box is moved
            pos_offset = item.boundingRect().topLeft()
            to_send_rect = new_rect.translated(- pos_offset)
            self.add_box_to_animation(
                item, int(to_send_rect.left()), int(to_send_rect.top()))

    def full_repulse(self):
        self._full_repulse_boxes.clear()
        for box in canvas.list_boxes():
            if box not in self._full_repulse_boxes:            
                self.deplace_boxes_from_repulsers([box])
        self._full_repulse_boxes.clear()

    def bring_neighbors_and_deplace_boxes(
            self, box_widget: BoxWidget, new_scene_rect: QRectF):
        neighbors = [box_widget]
        limit_top = box_widget.pos().y()
        
        for neighbor in neighbors:
            srect = neighbor.boundingRect()
            for moving_box in self.move_boxes:
                if moving_box.widget is neighbor:
                    srect.translate(moving_box.to_pt)
                    break
            else:
                srect.translate(neighbor.pos())

            for item in self.items(
                    srect.adjusted(0, 0, 0,
                                   canvas.theme.box_spacing + 1)):
                if item not in neighbors and isinstance(item, BoxWidget):
                    nrect = item.boundingRect().translated(item.pos())
                    if nrect.top() >= limit_top:
                        neighbors.append(item)
        
        neighbors.remove(box_widget)
        
        less_y = box_widget.boundingRect().height() - new_scene_rect.height()

        repulser_boxes = list[BoxWidget]()

        for neighbor in neighbors:
            self.add_box_to_animation(
                neighbor, int(neighbor.pos().x()), int(neighbor.pos().y() - less_y))
            repulser_boxes.append(neighbor)
        repulser_boxes.append(box_widget)
        
        self.deplace_boxes_from_repulsers(
            repulser_boxes, wanted_direction=Direction.UP)